# -*- coding: utf-8 -*-
#
# Copyright (C) 2011 Edgewall Software
# Copyright (C) 2011 Dirk Stöcker <trac@dstoecker.de>
# All rights reserved.
#
# This software is licensed as described in the file COPYING, which
# you should have received as part of this distribution. The terms
# are also available at http://trac.edgewall.com/license.html.
#
# This software consists of voluntary contributions made by many
# individuals. For the exact contribution history, see the revision
# history and logs, available at http://projects.edgewall.com/trac/.
#
# Author: Dirk Stöcker <trac@dstoecker.de>

import httplib
from xmlrpclib import ServerProxy, Transport


class TimeoutHTTPConnection(httplib.HTTPConnection):
    def __init__(self,host,timeout=3):
        httplib.HTTPConnection.__init__(self,host,timeout=timeout)

    # in python 2.6, xmlrpclib expects to use an httplib.HTTP
    # instance and will call getreply() and getfile()

    def getreply(self):
        response = self.getresponse()
        self.file = response.fp
        return response.status, response.reason, response.msg

    def getfile(self):
        return self.file


class TimeoutTransport(Transport):
    def __init__(self, timeout=3, *l, **kw):
        Transport.__init__(self,*l,**kw)
        self.timeout=timeout

    def make_connection(self, host):
        conn = TimeoutHTTPConnection(host,self.timeout)
        return conn


class TimeoutServerProxy(ServerProxy):
    def __init__(self,uri,timeout=3,*l,**kw):
        kw['transport']=TimeoutTransport(timeout=timeout, use_datetime=kw.get('use_datetime',0))
        ServerProxy.__init__(self,uri,*l,**kw)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2015 Edgewall Software
# Copyright (C) 2015 Dirk Stöcker <trac@dstoecker.de>
# All rights reserved.
#
# This software is licensed as described in the file COPYING, which
# you should have received as part of this distribution. The terms
# are also available at http://trac.edgewall.com/license.html.
#
# This software consists of voluntary contributions made by many
# individuals. For the exact contribution history, see the revision
# history and logs, available at http://projects.edgewall.com/trac/.
#
# Author: Dirk Stöcker <trac@dstoecker.de>

import urllib2
from xml.etree import ElementTree

from trac.core import Component, implements
from trac.util.html import html

from tracspamfilter.captcha import ICaptchaMethod
from tracspamfilter.filters.mollom import MollomFilterStrategy


class MollomCaptcha(Component):
    """An image captcha server by http://www.mollom.com/"""

    implements(ICaptchaMethod)

    def __init__(self):
        self.mollom = MollomFilterStrategy(self.env)

    def generate_captcha(self, req):
        resp, content = self.mollom._call('captcha', {'type': 'image'})
        tree = ElementTree.fromstring(content)
        req.session['captcha_mollom_id'] = tree.find('./captcha/id').text
        url = tree.find('./captcha/url').text
        return '', html.img(src=url, alt='captcha')

    def verify_captcha(self, req):
        mollom_id = req.session.get('captcha_mollom_id')
        solution = req.args.get('captcha_response', '')
        try:
            resp, content = self.mollom._call('captcha/%s' % mollom_id,
                                              {'solution': solution})
        except Exception, e:
            self.log.warning("Exception in Mollom captcha handling (%s)", e)
            return False
        else:
            tree = ElementTree.fromstring(content)
            del req.session['captcha_mollom_id']
            if tree.find('./captcha/solved').text == '1':
                return True
            else:
                self.log.warning("Mollom captcha returned error: %s",
                                 tree.find('./message'))
        return False

    def is_usable(self, req):
        try:
            return self.mollom.verify_key(req)
        except urllib2.URLError:
            return False



# -*- coding: utf-8 -*-
#
# Copyright (C) 2006 Edgewall Software
# All rights reserved.
#
# This software is licensed as described in the file COPYING, which
# you should have received as part of this distribution. The terms
# are also available at http://trac.edgewall.com/license.html.
#
# This software consists of voluntary contributions made by many
# individuals. For the exact contribution history, see the revision
# history and logs, available at http://projects.edgewall.com/trac/.

from __future__ import with_statement

import time
import unittest

from trac.core import Component, implements
from trac.test import EnvironmentStub, Mock

from tracspamfilter.api import IFilterStrategy, RejectContent
from tracspamfilter.filtersystem import FilterSystem
from tracspamfilter.model import LogEntry
from tracspamfilter.tests.model import drop_tables


class DummyStrategy(Component):
    implements(IFilterStrategy)

    def __init__(self):
        self.test_called = self.train_called = False
        self.req = self.author = self.content = None
        self.karma = 0
        self.message = None
        self.spam = None

    def configure(self, karma, message="Dummy"):
        self.karma = karma
        self.message = message

    def test(self, req, author, content, ip):
        self.test_called = True
        self.req = req
        self.author = author
        self.content = content
        self.ip = ip
        return self.karma, self.message

    def train(self, req, author, content, ip, spam=True):
        self.train_called = True
        self.req = req
        self.author = author
        self.content = content
        self.ip = ip
        self.spam = spam

    def is_external(self):
        return False

class FilterSystemTestCase(unittest.TestCase):

    def setUp(self):
        self.env = EnvironmentStub(enable=[FilterSystem, DummyStrategy])
        with self.env.db_transaction as db:
            FilterSystem(self.env).upgrade_environment(db)

    def tearDown(self):
        drop_tables(self.env)
        self.env.reset_db()

    def test_trust_authenticated(self):
        self.env.config.set('spam-filter', 'trust_authenticated', True)
        req = Mock(environ={}, path_info='/foo', authname='john',
                   args={}, remote_addr='127.0.0.1')
        FilterSystem(self.env).test(req, '', [])
        self.assertFalse(DummyStrategy(self.env).test_called)

    def test_dont_trust_authenticated(self):
        self.env.config.set('spam-filter', 'trust_authenticated', False)
        req = Mock(environ={}, path_info='/foo', authname='john',
                   args={}, remote_addr='127.0.0.1')
        FilterSystem(self.env).test(req, '', [])
        self.assertTrue(DummyStrategy(self.env).test_called)

    def test_without_oldcontent(self):
        req = Mock(environ={}, path_info='/foo', authname='anonymous',
                   args={}, remote_addr='127.0.0.1')
        FilterSystem(self.env).test(req, 'John Doe', [(None, 'Test')])
        self.assertEqual('Test', DummyStrategy(self.env).content)

    def test_with_oldcontent(self):
        req = Mock(environ={}, path_info='/foo', authname='anonymous',
                   args={}, remote_addr='127.0.0.1')
        FilterSystem(self.env).test(req, 'John Doe', [('Test', 'Test 1 2 3')])
        self.assertEqual('Test 1 2 3', DummyStrategy(self.env).content)

    def test_with_oldcontent_multiline(self):
        req = Mock(environ={}, path_info='/foo', authname='anonymous',
                   args={}, remote_addr='127.0.0.1')
        FilterSystem(self.env).test(req, 'John Doe', [('Text\n1 2 3\n7 8 9',
                                                       'Test\n1 2 3\n4 5 6')])
        self.assertEqual('Test\n4 5 6', DummyStrategy(self.env).content)

    def test_bad_karma(self):
        req = Mock(environ={}, path_info='/foo', authname='anonymous',
                   args={}, remote_addr='127.0.0.1')
        DummyStrategy(self.env).configure(-5, 'Blacklisted')
        try:
            FilterSystem(self.env).test(req, 'John Doe', [(None, 'Test')])
            self.fail('Expected RejectContent exception')
        except RejectContent, e:
            self.assertEqual('<div class="message">'
                             'Submission rejected as potential spam '
                             '<ul><li>Blacklisted</li></ul></div>', str(e))

    def test_good_karma(self):
        req = Mock(environ={}, path_info='/foo', authname='anonymous',
                   args={}, remote_addr='127.0.0.1')
        DummyStrategy(self.env).configure(5)
        FilterSystem(self.env).test(req, 'John Doe', [(None, 'Test')])

    def test_log_reject(self):
        req = Mock(environ={}, path_info='/foo', authname='anonymous',
                   args={}, remote_addr='127.0.0.1')
        DummyStrategy(self.env).configure(-5, 'Blacklisted')
        try:
            FilterSystem(self.env).test(req, 'John Doe', [(None, 'Test')])
            self.fail('Expected RejectContent exception')
        except RejectContent, e:
            pass

        log = list(LogEntry.select(self.env))
        self.assertEqual(1, len(log))
        entry = log[0]
        self.assertEqual('/foo', entry.path)
        self.assertEqual('John Doe', entry.author)
        self.assertEqual(False, entry.authenticated)
        self.assertEqual('127.0.0.1', entry.ipnr)
        self.assertEqual('Test', entry.content)
        self.assertEqual(True, entry.rejected)
        self.assertEqual(-5, entry.karma)
        self.assertEqual([['DummyStrategy', '-5', 'Blacklisted']], entry.reasons)

    def test_log_accept(self):
        req = Mock(environ={}, path_info='/foo', authname='anonymous',
                   args={}, remote_addr='127.0.0.1')
        DummyStrategy(self.env).configure(5)
        FilterSystem(self.env).test(req, 'John Doe', [(None, 'Test')])

        log = list(LogEntry.select(self.env))
        self.assertEqual(1, len(log))
        entry = log[0]
        self.assertEqual('/foo', entry.path)
        self.assertEqual('John Doe', entry.author)
        self.assertEqual(False, entry.authenticated)
        self.assertEqual('127.0.0.1', entry.ipnr)
        self.assertEqual('Test', entry.content)
        self.assertEqual(False, entry.rejected)
        self.assertEqual(5, entry.karma)
        self.assertEqual([['DummyStrategy', '5', 'Dummy']], entry.reasons)

    def test_train_spam(self):
        req = Mock(environ={'SERVER_NAME': 'localhost', 'SERVER_PORT': '80',
                            'wsgi.url_scheme': 'http'},
                   path_info='/foo', authname='anonymous',
                   args={}, remote_addr='127.0.0.1')
        entry = LogEntry(self.env, time.time(), '/foo', 'john', False,
                         '127.0.0.1', '', 'Test', False, 5, [], req)
        entry.insert()

        FilterSystem(self.env).train(req, entry.id, spam=True)

        strategy = DummyStrategy(self.env)
        self.assertEqual(True, strategy.train_called)
        self.assertEqual('john', strategy.author)
        self.assertEqual('Test', strategy.content)
        self.assertEqual(True, strategy.spam)

        log = list(LogEntry.select(self.env))
        self.assertEqual(1, len(log))
        entry = log[0]
        self.assertEqual(True, entry.rejected)

    def test_train_ham(self):
        req = Mock(environ={'SERVER_NAME': 'localhost', 'SERVER_PORT': '80',
                            'wsgi.url_scheme': 'http'},
                   path_info='/foo', authname='anonymous',
                   remote_addr='127.0.0.1')
        entry = LogEntry(self.env, time.time(), '/foo', 'john', False,
                         '127.0.0.1', '', 'Test', True, -5, [], req)
        entry.insert()
        FilterSystem(self.env).train(req, entry.id, spam=False)

        strategy = DummyStrategy(self.env)
        self.assertEqual(True, strategy.train_called)
        self.assertEqual('john', strategy.author)
        self.assertEqual('Test', strategy.content)
        self.assertEqual(False, strategy.spam)

        log = list(LogEntry.select(self.env))
        self.assertEqual(1, len(log))
        entry = log[0]
        self.assertEqual(False, entry.rejected)


def test_suite():
    suite = unittest.TestSuite()
    suite.addTest(unittest.makeSuite(FilterSystemTestCase))
    return suite

if __name__ == '__main__':
    unittest.main(defaultTest='test_suite')



# -*- coding: utf-8 -*-
#
# Copyright (C) 2006 Edgewall Software
# All rights reserved.
#
# This software is licensed as described in the file COPYING, which
# you should have received as part of this distribution. The terms
# are also available at http://trac.edgewall.com/license.html.
#
# This software consists of voluntary contributions made by many
# individuals. For the exact contribution history, see the revision
# history and logs, available at http://projects.edgewall.com/trac/.

from trac.db import Column, DatabaseManager, Table


def _schema_to_sql(env, db, table):
    connector, _ = DatabaseManager(env)._get_connector()
    return connector.to_sql(table)


def add_log_table(env, db):
    """Add a table for storing the spamfilter logs."""
    table = Table('spamfilter_log', key='id')[
        Column('id', auto_increment=True),
        Column('time', type='int'),
        Column('path'),
        Column('author'),
        Column('authenticated', type='int'),
        Column('ipnr'),
        Column('content'),
        Column('rejected', type='int'),
        Column('karma', type='int'),
        Column('reasons')
    ]
    cursor = db.cursor()
    for stmt in _schema_to_sql(env, db, table):
        cursor.execute(stmt)


def add_headers_column_to_log_table(env, db):
    """Add a column to the log table for storing the request headers."""
    table = Table('spamfilter_log', key='id')[
        Column('id', auto_increment=True),
        Column('time', type='int'),
        Column('path'),
        Column('author'),
        Column('authenticated', type='int'),
        Column('ipnr'),
        Column('headers'),
        Column('content'),
        Column('rejected', type='int'),
        Column('karma', type='int'),
        Column('reasons')
    ]
    cursor = db.cursor()
    cursor.execute("CREATE TEMPORARY TABLE spamfilter_log_old AS "
                   "SELECT * FROM spamfilter_log")
    cursor.execute("DROP TABLE spamfilter_log")
    for stmt in _schema_to_sql(env, db, table):
        cursor.execute(stmt)
    cursor.execute("INSERT INTO spamfilter_log (id,time,path,author,"
                   "authenticated,ipnr,content,rejected,karma,reasons) "
                   "SELECT id,time,path,author,authenticated,ipnr,content,"
                   "rejected,karma,reasons FROM spamfilter_log_old")
    cursor.execute("DROP TABLE spamfilter_log_old")


def add_bayes_table(env, db):
    """Add table required for bayesian filtering."""
    table = Table('spamfilter_bayes', key='word')[
        Column('word'),
        Column('nspam', type='int'),
        Column('nham', type='int')
    ]
    cursor = db.cursor()
    for stmt in _schema_to_sql(env, db, table):
        cursor.execute(stmt)


def add_statistics_table(env, db):
    """Add table required for filtering statistics."""
    table = Table('spamfilter_statistics',
                  key=['strategy', 'action', 'data', 'status'])[
        Column('strategy'),
        Column('action'),
        Column('data'),
        Column('status'),
        Column('delay', type='double precision'),
        Column('delay_max', type='double precision'),
        Column('delay_min', type='double precision'),
        Column('count', type='int'),
        Column('external', type='int'),
        Column('time', type='int')
    ]
    cursor = db.cursor()
    for stmt in _schema_to_sql(env, db, table):
        cursor.execute(stmt)


def add_request_column_to_log_table(env, db):
    """Add a column to the log table for storing the complete request."""
    table = Table('spamfilter_log', key='id')[
        Column('id', auto_increment=True),
        Column('time', type='int'),
        Column('path'),
        Column('author'),
        Column('authenticated', type='int'),
        Column('ipnr'),
        Column('headers'),
        Column('content'),
        Column('rejected', type='int'),
        Column('karma', type='int'),
        Column('reasons'),
        Column('request')
    ]
    cursor = db.cursor()
    cursor.execute("CREATE TEMPORARY TABLE spamfilter_log_old AS "
                   "SELECT * FROM spamfilter_log")
    cursor.execute("DROP TABLE spamfilter_log")
    for stmt in _schema_to_sql(env, db, table):
        cursor.execute(stmt)
    # copy old, but strip entries with non-XML reasons
    cursor.execute("INSERT INTO spamfilter_log (id,time,path,author,"
                   "authenticated,ipnr,headers,content,rejected,karma,reasons) "
                   "SELECT id,time,path,author,authenticated,ipnr,headers,content,"
                   "rejected,karma,reasons FROM spamfilter_log_old "
                   "WHERE reasons LIKE '<entries>%</entries>'")
    cursor.execute("DROP TABLE spamfilter_log_old")


def add_report_table(env, db):
    """Add table required for spam reports by users."""
    table = Table('spamfilter_report', key=['id'])[
        Column('id', auto_increment=True),
        Column('entry'),
        Column('headers'),
        Column('author'),
        Column('authenticated', type='int'),
        Column('comment'),
        Column('time', type='int')
    ]
    cursor = db.cursor()
    for stmt in _schema_to_sql(env, db, table):
        cursor.execute(stmt)

version_map = {
    1: [add_log_table],
    2: [add_headers_column_to_log_table],
    3: [add_bayes_table],
    4: [add_statistics_table, add_request_column_to_log_table, add_report_table]
}



# -*- coding: utf-8 -*-
#
# Copyright (C) 2015 Edgewall Software
# Copyright (C) 2015 Dirk Stöcker <trac@dstoecker.de>
# All rights reserved.
#
# This software is licensed as described in the file COPYING, which
# you should have received as part of this distribution. The terms
# are also available at http://trac.edgewall.com/license.html.
#
# This software consists of voluntary contributions made by many
# individuals. For the exact contribution history, see the revision
# history and logs, available at http://projects.edgewall.com/trac/.
#
# Author: Dirk Stöcker <trac@dstoecker.de>

import re
from dns.name import from_text
from dns.resolver import NXDOMAIN, NoAnswer, NoNameservers, Timeout, query

from trac.config import IntOption, ListOption
from trac.core import Component, implements

from tracspamfilter.api import IFilterStrategy, N_


class URLBlacklistFilterStrategy(Component):
    """Spam filter based on URL blacklistings.

    Requires the dnspython module from http://www.dnspython.org/.
    """
    implements(IFilterStrategy)

    karma_points = IntOption('spam-filter', 'url_blacklist_karma', '3',
        """By how many points blacklisting by a single bad URL impacts the
        overall karma of a submission.""", doc_domain='tracspamfilter')

    servers_default = 'urired.spameatingmonkey.net, multi.surbl.org, ' \
                      'dbl.spamhaus.org'
    servers = ListOption('spam-filter', 'url_blacklist_servers',
        servers_default, doc="Servers used for URL blacklisting.",
        doc_domain='tracspamfilter')

    # IFilterStrategy implementation

    def is_external(self):
        return True

    def test(self, req, author, content, ip):
        if not self._check_preconditions(req, author, content, ip):
            return

        urls = self._geturls(author + "\n" + content)

        if not urls:
            return

        self.log.debug('Checking for URL blacklisting on "%s"',
                       ", ".join(urls))

        points = 0
        servers = []

        for server in self.servers:
            for url in sorted(urls.keys()):
                self.log.debug("Checking blacklist %s for %s", server, url)
                try:
                    servers.append(self._query(url, server))
                    points -= abs(self.karma_points)
                except NXDOMAIN:  # not blacklisted on this server
                    #if url.startswith("www.") and not url[4:] in urls:
                    #    try:
                    #        self.log.debug("Checking blacklist %s for %s",
                    #                       server, url[4:])
                    #        servers.append("[www.]%s" % self._query(url[4:], server))
                    #        points -= abs(self.karma_points)
                    #    except:
                    #        pass
                    continue
                except (Timeout, NoAnswer, NoNameservers), e:
                    self.log.warning('Error checking URL blacklist server '
                                     '"%s" for URL "%s": %s', server, url, e)

        if points != 0:
            return points, N_("URL's blacklisted by %s"), ', '.join(servers)

    def train(self, req, author, content, ip, spam=True):
        return 0

    # Internal methods

    def _query(self, url, server):
        res = query(from_text(url + '.' + server.encode('utf-8')))[0].to_text()
        if res == '127.0.0.1':
            return '%s (%s)' % (server, url)
        # strip the common part of responses
        if res.startswith('127.0.0.'):
            res = res[8:]
        elif res.startswith('127.'):
            res = res[4:]
        return '%s (%s[%s])' % (server, url, res)

    def _check_preconditions(self, req, author, content, ip):
        if self.karma_points == 0 or not self.servers:
            return False

        return True

    def _geturls(self, content):
        urls = {}
        content = content.lower()
        # no IDN domains, only punnycode
        urlstr = re.compile("^([a-z0-9][a-z0-9.-]+[a-z0-9])(.?)")

        while 1:
            pos = content.find('//')
            if pos < 0:
                break
            content = content[pos + 2:]
            res = urlstr.search(content)
            if res:
                u = res.group(1)
                urls[u] = urls.get(u, 0)
                if res.group(2) not in ('"', '\'', '/', '\n', '.', '!',
                                        '?', ',', ';', ''):
                    self.log.warn("Strange URL '%s' found.", u)
        return urls



# -*- coding: utf-8 -*-
#
# Copyright (C) 2014 Dirk Stöcker <trac@dstoecker.de>
# All rights reserved.
#
# This software is licensed as described in the file COPYING, which
# you should have received as part of this distribution. The terms
# are also available at http://trac.edgewall.com/license.html.
#
# This software consists of voluntary contributions made by many
# individuals. For the exact contribution history, see the revision
# history and logs, available at http://projects.edgewall.com/trac/.

import httplib2
import oauth2
import urllib2
from email.Utils import parseaddr
from pkg_resources import get_distribution
from urllib import urlencode
from xml.etree import ElementTree

from trac import __version__ as TRAC_VERSION
from trac.config import IntOption, Option
from trac.core import Component, implements
from trac.env import ISystemInfoProvider
from trac.mimeview.api import is_binary
from trac.util import get_pkginfo

from tracspamfilter.api import IFilterStrategy, N_


class MollomFilterStrategy(Component):
    """Spam filter using the Mollom service (http://mollom.com/).
    """
    implements(IFilterStrategy, ISystemInfoProvider)

    karma_points = IntOption('spam-filter', 'mollom_karma', '5',
        """By how many points an Mollom reject impacts the overall karma
        of a submission.""", doc_domain ='tracspamfilter')

    public_key = Option('spam-filter', 'mollom_public_key', '',
        "Public key required to use the Mollom API.",
        doc_domain ='tracspamfilter')

    private_key = Option('spam-filter', 'mollom_private_key', '',
        "Private key required to use the Mollom API.",
        doc_domain ='tracspamfilter')

    api_url = Option('spam-filter', 'mollom_api_url', 'rest.mollom.com/v1/',
        "URL of the Mollom service.", doc_domain ='tracspamfilter')

    user_agent = 'Trac/%s | SpamFilter/%s' % (
        TRAC_VERSION, get_distribution('TracSpamFilter').version
    )

    def __init__(self):
        self.verified_key = None

    # IFilterStrategy implementation

    def is_external(self):
        return True

    def test(self, req, author, content, ip):
        if not self._check_preconditions(req, author, content):
            return

        try:
            # Split up author into name and email, if possible
            author = author.encode('utf-8')
            author_name, author_email = parseaddr(author)
            if not author_name and not author_email:
                author_name = author
            elif not author_name and author_email.find('@') < 1:
                author_name = author
                author_email = None

            params = {
                'authorIp': ip,
                'postBody': content.encode('utf-8'),
                'checks': 'spam',
                'authorName': author_name
            }
            if author_email:
                params['authorMail'] = author_email

            resp, content = self._call('content', params)
            if '<spamClassification>spam</spamClassification>' in content:
                tree = ElementTree.fromstring(content)
                confidence = 1
                se = tree.find('./content/spamScore')
                if se is not None:
                    confidence = float(se.text)
                else:
                    self.log.warn("Mollom score not found")
                karma = abs(self.karma_points) * float(confidence)
                self.log.debug("Mollom says content is %s spam", confidence)
                return -int(karma + 0.5), N_("Mollom says content is spam")
            elif '<spamClassification>ham</spamClassification>' in content:
                tree = ElementTree.fromstring(content)
                confidence = 1
                se = tree.find('./content/spamScore')
                if se is not None:
                    confidence = 1.0 - float(se.text)
                else:
                    self.log.warn("Mollom score not found")
                karma = abs(self.karma_points) * float(confidence)
                self.log.debug("Mollom says content is %s ham", confidence)
                return int(karma + 0.5), N_("Mollom says content is ham")

        except urllib2.URLError, e:
            self.log.warn("Mollom request failed (%s)", e)

    def train(self, req, author, content, ip, spam=True):
        return 0

    # ISystemInfoProvider methods

    def get_system_info(self):
        yield 'python-oauth2', get_pkginfo(oauth2)['version']
        yield 'httplib2', get_pkginfo(httplib2)['version']

    # Internal methods

    def _call(self, url, params=None, public_key=None, private_key=None,
              api_url=None):
        if not api_url:
            api_url = self.api_url
        if not public_key:
            public_key = self.public_key
        if not private_key:
            private_key = self.private_key
        headers = {
            'Content-Type': 'text/plain',
            'User-Agent': self.user_agent
        }
        if params:
            body = urlencode(params)
        else:
            body = '\n'

        url = 'http://' + api_url + url
        consumer = oauth2.Consumer(public_key, private_key)
        req = oauth2.Request.from_consumer_and_token(
            consumer, http_method='POST', http_url=url, body=body)
        req.sign_request(oauth2.SignatureMethod_HMAC_SHA1(), consumer, None)
        headers.update(req.to_header())
        return httplib2.Http().request(url, method="POST", body=body,
                                       headers=headers)

    def _check_preconditions(self, req, author, content):
        if self.karma_points == 0:
            return False

        if not self.public_key or not self.private_key:
            self.log.warning("Mollom API keys missing")
            return False

        if is_binary(content):
            self.log.warning("Content is binary, Mollom content check "
                             "skipped")
            return False

        try:
            if not self.verify_key(req):
                self.log.warning("Mollom API keys are invalid")
                return False
            return True
        except urllib2.URLError, e:
            self.log.warn("Mollom request failed (%s)", e)

    def verify_key(self, req, api_url=None, public_key=None, private_key=None):
        if api_url is None:
            api_url = self.api_url
        if private_key is None:
            private_key = self.private_key
        if public_key is None:
            public_key = self.public_key

        if public_key + private_key != self.verified_key:
            self.log.debug("Verifying Mollom API keys")
            resp, content = self._call('site/%s' % public_key, None,
                                       public_key, private_key, api_url)
            c = "<privateKey>%s</privateKey>" % private_key
            if c in content:
                self.log.debug("Mollom API keys are valid")
                self.verified = True
                self.verified_key = public_key + private_key

        return self.verified_key is not None



# -*- coding: utf-8 -*-
#
# Copyright (C) 2015 Edgewall Software
# Copyright (C) 2006 Matthew Good <trac@matt-good.net>
# Copyright (C) 2009 Vaclav Slavik <vslavik@fastmail.fm>
# Copyright (C) 2015 Dirk Stöcker <trac@dstoecker.de>
# All rights reserved.
#
# This software is licensed as described in the file COPYING, which
# you should have received as part of this distribution. The terms
# are also available at http://trac.edgewall.com/license.html.
#
# This software consists of voluntary contributions made by many
# individuals. For the exact contribution history, see the revision
# history and logs, available at http://projects.edgewall.com/trac/.
#
# Author: Vaclav Slavik <vslavik@fastmail.fm>,
#         Matthew Good <trac@matt-good.net>

from dns.resolver import NXDOMAIN, NoAnswer, NoNameservers, Timeout, query

from trac.config import Option, IntOption
from trac.core import Component, implements

from tracspamfilter.api import IFilterStrategy, N_


class HttpBLFilterStrategy(Component):
    """Spam filter based on Project Honey Pot's Http:BL blacklist.

    Requires the dnspython module from http://www.dnspython.org/.
    """
    implements(IFilterStrategy)

    karma_points = IntOption('spam-filter', 'httpbl_spammer_karma', '6',
        """By how many points listing as "comment spammer" impacts the
        overall karma of a submission.""", doc_domain='tracspamfilter')

    api_key = Option('spam-filter', 'httpbl_api_key', '',
        "Http:BL API key required for use.", doc_domain='tracspamfilter')

    # IFilterStrategy implementation

    def is_external(self):
        return True

    def test(self, req, author, content, ip):
        if not self.api_key:
            self.log.warning("API key not configured.")
            return

        # IPV4 address? HTTP:BL does not yet support IPv6
        if ip.find(".") < 0:
            return

        reverse_octal = '.'.join(reversed(ip.split('.')))
        addr = '%s.%s.dnsbl.httpbl.org' % (self.api_key, reverse_octal)
        self.log.debug('Querying Http:BL: %s', addr)

        try:
            dns_answer = query(addr)
            answer = [int(i) for i in str(dns_answer[0]).split('.')]
            if answer[0] != 127:
                self.log.warning('Invalid Http:BL reply for IP "%s": %s',
                                 ip, dns_answer)
                return

            # TODO: answer[1] represents number of days since last activity
            #       and answer[2] is treat score assigned by Project Honey
            #       Pot. We could use both to adjust karma.

            is_suspicious = answer[3] & 1
            is_spammer = answer[3] & 4

            points = 0
            if is_suspicious:
                points -= abs(self.karma_points) / 3
            if is_spammer:
                points -= abs(self.karma_points)

            if points != 0:
                return points, N_("IP %s blacklisted by Http:BL"), ip

        except NXDOMAIN:
            # not blacklisted on this server
            return
        except (Timeout, NoAnswer, NoNameservers), e:
            self.log.warning('Error checking Http:BL for IP "%s": %s', ip, e)

    def train(self, req, author, content, ip, spam=True):
        return 0



# -*- coding: utf-8 -*-
#
# Copyright (C) 2006 Edgewall Software
# All rights reserved.
#
# This software is licensed as described in the file COPYING, which
# you should have received as part of this distribution. The terms
# are also available at http://trac.edgewall.com/license.html.
#
# This software consists of voluntary contributions made by many
# individuals. For the exact contribution history, see the revision
# history and logs, available at http://projects.edgewall.com/trac/.

from __future__ import with_statement

import time
import unittest
from datetime import datetime, timedelta

from trac.test import EnvironmentStub

from tracspamfilter.filtersystem import FilterSystem
from tracspamfilter.model import LogEntry


def drop_tables(env):
    with env.db_transaction as db:
        for table in ('spamfilter_bayes', 'spamfilter_log',
                      'spamfilter_report', 'spamfilter_statistics'):
            db("DROP TABLE IF EXISTS %s" % table)
        db("DELETE FROM system WHERE name='spamfilter_version'")


class LogEntryTestCase(unittest.TestCase):

    def setUp(self):
        self.env = EnvironmentStub()
        with self.env.db_transaction as db:
            FilterSystem(self.env).upgrade_environment(db)

    def tearDown(self):
        drop_tables(self.env)
        self.env.reset_db()

    def test_purge(self):
        now = datetime.now()
        oneweekago = time.mktime((now - timedelta(weeks=1)).timetuple())
        onedayago = time.mktime((now - timedelta(days=1)).timetuple())
        req = None

        LogEntry(self.env, oneweekago, '/foo', 'john', False, '127.0.0.1',
                 '', 'Test', False, 5, [], req).insert()
        LogEntry(self.env, onedayago, '/foo', 'anonymous', False, '127.0.0.1',
                 '', 'Test', True, -3, [], req).insert()

        LogEntry.purge(self.env, days=4)

        log = list(LogEntry.select(self.env))
        self.assertEqual(1, len(log))
        entry = log[0]
        self.assertEqual('anonymous', entry.author)


def test_suite():
    suite = unittest.TestSuite()
    suite.addTest(unittest.makeSuite(LogEntryTestCase))
    return suite

if __name__ == '__main__':
    unittest.main(defaultTest='test_suite')



# -*- coding: utf-8 -*-
#
# Copyright (C) 2015 Dirk Stöcker <trac@dstoecker.de>
# All rights reserved.
#
# This software is licensed as described in the file COPYING, which
# you should have received as part of this distribution. The terms
# are also available at http://trac.edgewall.com/license.html.
#
# This software consists of voluntary contributions made by many
# individuals. For the exact contribution history, see the revision
# history and logs, available at http://projects.edgewall.com/trac/.

import hashlib
import random
import urllib2
from pkg_resources import get_distribution

from trac import __version__ as TRAC_VERSION
from trac.config import Option
from trac.core import Component, implements
from trac.util.html import tag

from tracspamfilter.captcha import ICaptchaMethod


class KeycaptchaCaptcha(Component):
    """KeyCaptcha implementation"""

    implements(ICaptchaMethod)

    private_key = Option('spam-filter', 'captcha_keycaptcha_private_key', '',
        """Private key for KeyCaptcha usage.""", doc_domain="tracspamfilter")

    user_id = Option('spam-filter', 'captcha_keycaptcha_user_id', '',
        """User id for KeyCaptcha usage.""", doc_domain="tracspamfilter")

    user_agent = 'Trac/%s | SpamFilter/%s' \
                 % (TRAC_VERSION, get_distribution('TracSpamFilter').version)

    def generate_captcha(self, req):
        session_id = "%d-3.4.0.001" % random.randint(1, 10000000)
        sign1 = hashlib.md5(session_id + req.remote_addr +
                            self.private_key).hexdigest()
        sign2 = hashlib.md5(session_id + self.private_key).hexdigest()
        varblock = "var s_s_c_user_id = '%s';\n" % self.user_id
        varblock += "var s_s_c_session_id = '%s';\n" % session_id
        varblock += "var s_s_c_captcha_field_id = 'keycaptcha_response_field';\n"
        varblock += "var s_s_c_submit_button_id = 'keycaptcha_response_button';\n"
        varblock += "var s_s_c_web_server_sign = '%s';\n" % sign1
        varblock += "var s_s_c_web_server_sign2 = '%s';\n" % sign2
        varblock += "document.s_s_c_debugmode=1;\n"
        fragment = tag(tag.script(varblock, type='text/javascript'))

        fragment.append(
            tag.script(type='text/javascript',
                       src='http://backs.keycaptcha.com/swfs/cap.js')
        )

        fragment.append(
            tag.input(type='hidden', id='keycaptcha_response_field',
                      name='keycaptcha_response_field')
        )

        fragment.append(
            tag.input(type='submit', id='keycaptcha_response_button',
                      name='keycaptcha_response_button')
        )

        req.session['captcha_key_session'] = session_id

        return None, fragment

    def verify_key(self, private_key, user_id):
        if private_key is None or user_id is None:
            return False
        # FIXME - Not yet implemented
        return True

    def verify_captcha(self, req):
        session = None
        if 'captcha_key_session' in req.session:
            session = req.session['captcha_key_session']
            del req.session['captcha_key_session']

        response_field = req.args.get('keycaptcha_response_field')
        val = response_field.split('|')
        s = hashlib.md5('accept' + val[1] + self.private_key +
                        val[2]).hexdigest()
        self.log.debug("KeyCaptcha response: %s .. %s .. %s",
                       response_field, s, session)
        if s == val[0] and session == val[3]:
            try:
                request = urllib2.Request(
                    url=val[2],
                    headers={"User-agent": self.user_agent}
                )

                response = urllib2.urlopen(request)
                return_values = response.read()
                response.close()
            except Exception, e:
                self.log.warning("Exception in KeyCaptcha handling (%s)", e)
            else:
                self.log.debug("KeyCaptcha check result: %s", return_values)
                if return_values == '1':
                    return True
                self.log.warning("KeyCaptcha returned invalid check result: "
                                 "%s (%s)", return_values, response_field)
        else:
            self.log.warning("KeyCaptcha returned invalid data: "
                             "%s (%s,%s)", response_field, s, session)
        return False

    def is_usable(self, req):
        return self.private_key and self.user_id



# -*- coding: utf-8 -*-
#
# Copyright (C) 2006 Edgewall Software
# All rights reserved.
#
# This software is licensed as described in the file COPYING, which
# you should have received as part of this distribution. The terms
# are also available at http://trac.edgewall.com/license.html.
#
# This software consists of voluntary contributions made by many
# individuals. For the exact contribution history, see the revision
# history and logs, available at http://projects.edgewall.com/trac/.

import StringIO
import unittest

from trac.test import EnvironmentStub, Mock

from tracspamfilter.filters import akismet
from tracspamfilter.filters.akismet import AkismetFilterStrategy


class DummyRequest(object):

    def __init__(self, url, params, headers):
        self.url = url
        self.params = params
        self.headers = headers

akismet.urllib2.Request = DummyRequest


class DummyURLOpener(object):

    def __init__(self):
        self.responses = []
        self.requests = []

    def __call__(self, request):
        self.requests.append(request)
        return StringIO.StringIO(self.responses.pop(0))


class AkismetFilterStrategyTestCase(unittest.TestCase):

    def setUp(self):
        self.env = EnvironmentStub(enable=[AkismetFilterStrategy])
        self.strategy = AkismetFilterStrategy(self.env)
        self.urlopen = akismet.urllib2.urlopen = DummyURLOpener()

    def test_no_api_key(self):
        req = Mock(remote_addr='127.0.0.1')
        retval = self.strategy.test(req, 'anonymous', 'foobar',
                                    req.remote_addr)
        self.assertEqual(None, retval)

    def test_bad_api_key(self):
        req = Mock(authname='anonymous', base_url='http://example.org/',
                   remote_addr='127.0.0.1')
        self.env.config.set('spam-filter', 'akismet_api_key', 'INVALID')

        self.urlopen.responses = ['invalid']
        retval = self.strategy.test(req, 'anonymous', 'foobar',
                                    req.remote_addr)
        self.assertEqual(None, retval)
        self.assertEqual(1, len(self.urlopen.requests))

        req = self.urlopen.requests[0]
        self.assertEqual('http://rest.akismet.com/1.1/verify-key', req.url)
        self.assertEqual('blog=http%3A%2F%2Fexample.org%2F&key=INVALID',
                         req.params)
        self.assertEqual(self.strategy.user_agent, req.headers['User-Agent'])

    def test_check_ham(self):
        req = Mock(authname='anonymous', base_url='http://example.org/',
                   remote_addr='127.0.0.1', get_header=lambda x: None,
                   environ={})
        self.env.config.set('spam-filter', 'akismet_api_key', 'mykey')

        self.urlopen.responses = ['valid', 'false']
        retval = self.strategy.test(req, 'anonymous', 'foobar',
                                    req.remote_addr)
        self.assertEqual(None, retval)
        self.assertEqual(2, len(self.urlopen.requests))

        req = self.urlopen.requests[1]
        self.assertEqual('http://mykey.rest.akismet.com/1.1/comment-check',
                         req.url)
        self.assertEqual('user_ip=127.0.0.1&comment_type=trac&'
                         'referrer=unknown&blog=http%3A%2F%2Fexample.org%2F&'
                         'user_agent=None&comment_content=foobar&'
                         'comment_author=anonymous', req.params)
        self.assertEqual(self.strategy.user_agent, req.headers['User-Agent'])

    def test_check_spam(self):
        req = Mock(authname='anonymous', base_url='http://example.org/',
                   remote_addr='127.0.0.1', get_header=lambda x: None,
                   environ={})
        self.env.config.set('spam-filter', 'akismet_api_key', 'mykey')

        self.urlopen.responses = ['valid', 'true']
        retval = self.strategy.test(req, 'anonymous', 'foobar',
                                    req.remote_addr)
        self.assertEqual((-10, 'Akismet says content is spam'), retval)
        self.assertEqual(2, len(self.urlopen.requests))

    def test_submit_ham(self):
        req = Mock(authname='anonymous', base_url='http://example.org/',
                   remote_addr='127.0.0.1', get_header=lambda x: None,
                   environ={})
        self.env.config.set('spam-filter', 'akismet_api_key', 'mykey')

        self.urlopen.responses = ['valid', '']
        self.strategy.train(req, 'anonymous', 'foobar', req.remote_addr,
                            spam=False)

        req = self.urlopen.requests[1]
        self.assertEqual('http://mykey.rest.akismet.com/1.1/submit-ham',
                         req.url)
        self.assertEqual('user_ip=127.0.0.1&comment_type=trac&'
                         'referrer=unknown&blog=http%3A%2F%2Fexample.org%2F&'
                         'user_agent=None&comment_content=foobar&'
                         'comment_author=anonymous', req.params)
        self.assertEqual(self.strategy.user_agent, req.headers['User-Agent'])

    def test_submit_spam(self):
        req = Mock(authname='anonymous', base_url='http://example.org/',
                   remote_addr='127.0.0.1', get_header=lambda x: None,
                   environ={})
        self.env.config.set('spam-filter', 'akismet_api_key', 'mykey')

        self.urlopen.responses = ['valid', '']
        self.strategy.train(req, 'anonymous', 'foobar', req.remote_addr,
                            spam=True)

        req = self.urlopen.requests[1]
        self.assertEqual('http://mykey.rest.akismet.com/1.1/submit-spam',
                         req.url)
        self.assertEqual('user_ip=127.0.0.1&comment_type=trac&'
                         'referrer=unknown&blog=http%3A%2F%2Fexample.org%2F&'
                         'user_agent=None&comment_content=foobar&'
                         'comment_author=anonymous',
                         req.params)
        self.assertEqual(self.strategy.user_agent, req.headers['User-Agent'])


def test_suite():
    suite = unittest.TestSuite()
    suite.addTest(unittest.makeSuite(AkismetFilterStrategyTestCase))
    return suite

if __name__ == '__main__':
    unittest.main(defaultTest='test_suite')



# -*- coding: utf-8 -*-
#
# Copyright (C) 2005-2012 Edgewall Software
# Copyright (C) 2005-2006 Matthew Good <trac@matt-good.net>
# Copyright (C) 2006 Christopher Lenz <cmlenz@gmx.de>
# All rights reserved.
#
# This software is licensed as described in the file COPYING, which
# you should have received as part of this distribution. The terms
# are also available at http://trac.edgewall.com/license.html.
#
# This software consists of voluntary contributions made by many
# individuals. For the exact contribution history, see the revision
# history and logs, available at http://projects.edgewall.com/trac/.
#
# Author: Matthew Good <trac@matt-good.net>
#         Christopher Lenz <cmlenz@gmx.de>

from trac.core import Interface, TracError
from trac.util.translation import domain_functions

_, tag_, N_, add_domain, gettext, ngettext = domain_functions(
    'tracspamfilter',
    ('_', 'tag_', 'N_', 'add_domain', 'gettext', 'ngettext'))


class RejectContent(TracError):
    """Exception raised when content is rejected by a filter."""


class IFilterStrategy(Interface):
    """Mainfilter class, mainly consisting of test() and train() function

       Filters using network access should return True for is_external() call.

       Any variable ending in "karma_points" is presented in the karma admin
       interface.
    """

    def is_external(self):
        """Is this an service sending data to external servers.

        Return True if data is passed to external servers
        """

    def test(req, author, content, ip):
        """Test the given content submission.

        Should return a `(points, reason)` tuple to affect the score of the
        submission, where `points` is an integer, and `reason` is a brief
        description of why the score is being affected.

        If the filter strategy does not want (or is not able) to effectively
        test the submission, it should return `None`.
        """

    def train(req, author, content, ip, spam=True):
        """Train the filter by reporting a false negative or positive.

        The spam keyword argument is `True` if the content should be considered
        spam (a false negative), and `False` if the content was legitimate (a
        false positive).

        returns
           0 (no training),
           1 (training ok),
          -1 (training error) or
          -2 (missing preconditions, e.g. keys)
        """


class IRejectHandler(Interface):
    """Handle content rejection."""

    def reject_content(req, reason):
        """Reject content. `reason` is a human readable message describing why
        the content was rejected. """


def get_strategy_name(strategy):
    try:
        name = strategy._name
    except Exception:
        name = strategy.__class__.__name__
    if name.endswith("FilterStrategy"):
        name = name[:-14]
    return name



# -*- coding: utf-8 -*-
#
# Copyright (C) 2006 Edgewall Software
# All rights reserved.
#
# This software is licensed as described in the file COPYING, which
# you should have received as part of this distribution. The terms
# are also available at http://trac.edgewall.com/license.html.
#
# This software consists of voluntary contributions made by many
# individuals. For the exact contribution history, see the revision
# history and logs, available at http://projects.edgewall.com/trac/.

import unittest

from trac.core import TracError
from trac.perm import PermissionSystem
from trac.test import EnvironmentStub, Mock
from trac.web.api import RequestDone
from trac.wiki.model import WikiPage
from trac.wiki.web_ui import WikiModule

from tracspamfilter.filters import regex
from tracspamfilter.filters.regex import RegexFilterStrategy
from tracspamfilter.tests.compat import MockRequest


class DummyWikiPage(object):

    def __init__(self):
        self.text = ''

    def __call__(self, env, name):
        self.env = env
        self.name = name
        self.exists = True
        return self


class RegexFilterStrategyTestCase(unittest.TestCase):

    def setUp(self):
        self.env = EnvironmentStub(enable=['trac.*', RegexFilterStrategy],
                                   default_data=True)
        self.page = regex.WikiPage = DummyWikiPage()
        self.strategy = RegexFilterStrategy(self.env)

    def tearDown(self):
        self.env.reset_db()

    def _dispatch_request(self, req):
        module = WikiModule(self.env)
        self.assertTrue(module.match_request(req))
        try:
            params = module.process_request(req)
        except RequestDone:
            return []
        else:
            return params

    def test_no_patterns(self):
        retval = self.strategy.test(Mock(), 'anonymous', 'foobar',
                                    '127.0.0.1')
        self.assertEqual(None, retval)

    def test_one_matching_pattern(self):
        self.page.text = """{{{
foobar
}}}"""
        self.strategy.wiki_page_changed(self.page)
        retval = self.strategy.test(Mock(), 'anonymous', 'foobar',
                                    '127.0.0.1')
        self.assertEqual((-5, "Content contained these blacklisted "
                              "patterns: %s", '\'foobar\''), retval)

    def test_multiple_matching_pattern(self):
        self.page.text = """{{{
foobar
^foo
bar$
}}}"""
        self.strategy.wiki_page_changed(self.page)
        retval = self.strategy.test(Mock(), 'anonymous', '\nfoobar',
                                    '127.0.0.1')
        self.assertEqual((-10, "Content contained these blacklisted "
                               "patterns: %s", '\'foobar\', \'bar$\''),
                         retval)

    def test_view_page_with_invalid_pattern(self):
        """Page with invalid pattern should render fine, but not allow
        an edit without correcting the invalid pattern.
        """
        text = """{{{
(?i)eventbrite\.com
(?i)sneaker(?:supplier|nice\.com
}}}"""
        page = WikiPage(self.env)
        page.text = text
        page.name = 'BadContent'
        try:
            page.save('anonymous', 'Page created.')
        except TracError:
            self.assertTrue(WikiPage(self.env, 'BadContent').exists)
        else:
            self.fail("Saving page with invalid content did not "
                      "raise a TracError.")
        req = MockRequest(self.env, authname='user', args={
            'action': 'view',
        }, path_info='/wiki/BadContent')

        data = self._dispatch_request(req)[1]

        self.assertEqual(page.text, data['text'])

        req = MockRequest(self.env, authname='user', args={
            'action': 'edit',
            'preview': True,
            'version': 1,
            'text': text
        }, method='POST', path_info='/wiki/BadContent')
        self._dispatch_request(req)

        self.assertIn('Invalid Wiki page: Error in pattern '
                      '<tt>(?i)sneaker(?:supplier|nice\\.com</tt>: '
                      '<i>unbalanced parenthesis</i>.',
                      req.chrome['warnings'])

    def test_save_page_with_invalid_pattern(self):
        """Page cannot be saved with an invalid pattern."""
        perm = PermissionSystem(self.env)
        perm.grant_permission('user', 'authenticated')
        text = """{{{
        (?i)eventbrite\.com
        (?i)sneaker(?:supplier|nice\.com
        }}}"""
        req = MockRequest(self.env, authname='user', args={
            'action': 'edit',
            'text': text,
            'version': 0,
        }, method='POST', path_info='/wiki/BadContent')

        self._dispatch_request(req)

        self.assertIn('Invalid Wiki page: Error in pattern '
                      '<tt>(?i)sneaker(?:supplier|nice\\.com</tt>: '
                      '<i>unbalanced parenthesis</i>.',
                      req.chrome['warnings'])

    def test_save_page_with_valid_patterns(self):
        """Page with valid patterns can be saved."""
        perm = PermissionSystem(self.env)
        perm.grant_permission('user', 'authenticated')
        text = """{{{
        (?i)eventbrite\.com
        (?i)sneaker(?:supplier|nice)\.com
        }}}"""
        req = MockRequest(self.env, authname='user', args={
            'action': 'edit',
            'text': text,
            'version': 0,
        }, method='POST', path_info='/wiki/BadContent')

        self._dispatch_request(req)

        self.assertIn('Your changes have been saved in version 1.',
                      req.chrome['notices'])



def test_suite():
    suite = unittest.TestSuite()
    suite.addTest(unittest.makeSuite(RegexFilterStrategyTestCase))
    return suite

if __name__ == '__main__':
    unittest.main(defaultTest='test_suite')



# -*- coding: utf-8 -*-
#
# Copyright (C) 2012 Dirk Stöcker <trac@dstoecker.de>
# All rights reserved.
#
# This software is licensed as described in the file COPYING, which
# you should have received as part of this distribution. The terms
# are also available at http://trac.edgewall.com/license.html.
#
# This software consists of voluntary contributions made by many
# individuals. For the exact contribution history, see the revision
# history and logs, available at http://projects.edgewall.com/trac/.

from genshi.filters.transform import Transformer
from trac.config import IntOption, Option
from trac.core import Component, implements
from trac.util.html import tag
from trac.util.text import javascript_quote, shorten_line
from trac.web.api import ITemplateStreamFilter

from tracspamfilter.api import IFilterStrategy, N_


class TrapFieldFilterStrategy(Component):
    """Spam filter using a hidden trap field.
    """
    implements(IFilterStrategy, ITemplateStreamFilter)

    karma_points = IntOption('spam-filter', 'trap_karma', '10',
        """By how many points a trap reject impacts the overall karma of
        a submission.""", doc_domain='tracspamfilter')

    name = Option('spam-filter', 'trap_name', 'sfp_email',
        """Name of the invisible trap field, should contain some reference
        to e-mail for better results.""", doc_domain='tracspamfilter')

    name_hidden = Option('spam-filter', 'trap_name_hidden', 'sfph_mail',
        """Name of the hidden trap field, should contain some reference
        to e-mail for better results.""", doc_domain='tracspamfilter')

    name_register = Option('spam-filter', 'trap_name_register', 'spf_homepage',
        """Name of the register trap field, should contain some reference
        to web/homepage for better results.""", doc_domain='tracspamfilter')

    # IFilterStrategy implementation

    def is_external(self):
        return False

    def get_trap(self, req):
        i = req.args.getfirst(self.name)
        h = req.args.getfirst(self.name_hidden)
        if i and h and i != h:
            return i + "\n" + h
        elif h:
            return h
        return i

    def test(self, req, author, content, ip):
        i = req.args.getfirst(self.name)
        h = req.args.getfirst(self.name_hidden)
        r = self.getlink(req)

        if i and h:
            i = shorten_line(javascript_quote(i), 50)
            h = shorten_line(javascript_quote(h), 50)
            return -abs(self.karma_points), \
                   N_("Both trap fields says this is spam (%s, %s)"), i, h
        elif i:
            i = shorten_line(javascript_quote(i), 50)
            return -abs(self.karma_points), \
                   N_("Invisible trap field says this is spam (%s)"), i
        elif h:
            h = shorten_line(javascript_quote(h), 50)
            return -abs(self.karma_points), \
                   N_("Hidden trap field says this is spam (%s)"), h
        elif r:
            r = shorten_line(javascript_quote(r), 50)
            return -abs(self.karma_points), \
                   N_("Register trap field starts with HTTP URL (%s)"), r

    def getlink(self, req):
        r = req.args.getfirst(self.name_register)
        if not (r and (r.startswith('http://') or r.startswith('https://'))):
            return None
        return r

    def train(self, req, author, content, ip, spam=True):
        return 0

    # ITemplateStreamFilter interface
    def filter_stream(self, req, method, filename, stream, data):
        if self.karma_points > 0:
            # Insert the hidden field right before the submit buttons
            trap = tag.div(style='display:none;')(
                    tag.input(type='text', name=self.name, value=''),
                    tag.input(type='hidden', name=self.name_hidden, value=''))
            stream = stream | \
                     Transformer('//div[@class="buttons"]').before(trap)
        return stream



# -*- coding: utf-8 -*-
#
# Copyright (C) 2006 Edgewall Software
# All rights reserved.
#
# This software is licensed as described in the file COPYING, which
# you should have received as part of this distribution. The terms
# are also available at http://trac.edgewall.com/license.html.
#
# This software consists of voluntary contributions made by many
# individuals. For the exact contribution history, see the revision
# history and logs, available at http://projects.edgewall.com/trac/.

from time import time
from urllib import quote, unquote

from trac.admin import IAdminPanelProvider
from trac.config import ChoiceOption, IntOption
from trac.core import Component, TracError, implements
from trac.web.chrome import add_notice, add_stylesheet

from tracspamfilter.api import _, ngettext
from tracspamfilter.users import UserInfo


class UserAdminPageProvider(Component):
    """Web administration panel for spam user info."""

    MIN_WIKI =  IntOption('spam-filter', 'spam_user_minwiki', '0',
    "How many wiki edits are still an unused account.",
    doc_domain='tracspamfilter')

    MAX_AGE =  IntOption('spam-filter', 'spam_user_maxage', '200',
    "How many days no login are considered for dead accounts.",
    doc_domain='tracspamfilter')

    modes = ['overview', 'unused', 'authorized', 'all']
    DEFAULT_MODE =  ChoiceOption('spam-filter', 'spam_user_defaultmode',
    modes, "Default mode for spam user admin panel.",
    doc_domain='tracspamfilter')

    implements(IAdminPanelProvider)

    # IAdminPanelProvider methods

    def get_admin_panels(self, req):
        if 'SPAM_USER' in req.perm:
            yield ('spamfilter', _('Spam Filtering'), 'user', _('Users'))

    def render_admin_panel(self, req, cat, page, path_info):
        req.perm.require('SPAM_USER')

        if req.method == 'POST':
            if 'cleantemp' in req.args:
                UserInfo.deletetemporary(self.env)
            elif 'changeuser' in req.args:
                if not 'userold' in req.args or not 'usernew' in req.args:
                    raise TracError(_('Old or new value cannot be empty'))
                old = req.args['userold']
                new = req.args['usernew']
                # for strange usernames entering names already encoded is helpful
                if req.args.get('encoded', 0):
                    old = unquote(old).decode('utf8')
                    new = unquote(new).decode('utf8')
                if old == new:
                    raise TracError(_('Old and new value cannot be equal'))
                res = UserInfo.changeuser(self.env, old, new, req.args.get('auth', ''))
                if res == -3:
                    raise TracError(_('New name cannot be used in CC fields'))
                elif res < 0:
                    raise TracError(_('Illegal user arguments passed or changing not allowed'))
                elif res:
                    add_notice(req, ngettext('%(num)d entry has been updated', '%(num)d entries have been updated', res))
            elif 'fixemails' in req.args:
                users, stats = UserInfo.getinfo(self.env, 'authorized')
                for name,user in sorted(users.iteritems()):
                    if user[3] and user[3] != name and ((user[4] | user[6]) & 2):
                        res = UserInfo.changeuser(self.env, user[3], name)
                        if res == -3:
                            add_notice(req, _('Username \'%s\' cannot be used in CC fields') % name)
                        elif res < 0:
                            add_notice(req, _('Error for e-mail change for username \'%s\'') % name)
                        elif res:
                            add_notice(req, ngettext('%(num)d entry has been updated for user %(user)s',
                            '%(num)d entries have been updated for user %(user)s', res, user=name))
                        elif res:
                            add_notice(req, _('E-mails for user %s updated') % name)

            req.redirect(req.href.admin(cat, page,
                                        mode=req.args.get('mode')))

        data = {}
        data['curtime'] = int(time())
        data['maxage'] = int(req.args.get('maxage', self.MAX_AGE))*24*60*60
        data['tempcount'] = len(UserInfo.gettemporary(self.env))
        data['accmgr'] = 'ACCTMGR_USER_ADMIN' in req.perm
        mode = req.args.get('mode', self.DEFAULT_MODE)
        data['mode'] = mode
        data['encoded'] = req.args.get('encoded','')
        data['auth'] = req.args.get('auth','')
        minwiki = int(req.args.get('minwiki', self.MIN_WIKI))
        if 'user' in req.args:
            user = req.args['user']
            data['username'] = user
            data['user'] = UserInfo.getuserinfo(self.env, user)
            data['users'] = []
            data['stats'] = None
        else:
            data['username'] = None
            data['user'] = []
            users, stats = UserInfo.getinfo(self.env, mode if mode != 'unusedmulti' else 'unused', minwiki)
            data['users'] = users
            data['stats'] = stats
        data['quote'] = quote
        if mode == 'overview':
            data['usertype'] = _('data overview')
        elif mode == 'unused' or mode == 'unusedmulti':
            data['usertype'] = _('unused accounts')
        elif mode == 'authorized':
            data['usertype'] = _('registered accounts')
        elif mode == 'user':
            data['usertype'] = _("detailed user information for '%s'") % data['username']
        else:
            data['usertype'] = _('everything from accounts, wiki, tickets and svn')
        data['entrytext'] = ngettext('%(num)d entry', '%(num)d entries', (stats['numtotal'] if mode=='overview' else len(data['users'])))

        add_stylesheet(req, 'spamfilter/admin.css')
        return 'admin_user.html', data



# -*- coding: utf-8 -*-
#
# Copyright (C) 2015 Dirk Stöcker <trac@dstoecker.de>
# All rights reserved.
#
# This software is licensed as described in the file COPYING, which
# you should have received as part of this distribution. The terms
# are also available at http://trac.edgewall.com/license.html.
#
# This software consists of voluntary contributions made by many
# individuals. For the exact contribution history, see the revision
# history and logs, available at http://projects.edgewall.com/trac/.

import urllib
import urllib2
from pkg_resources import get_distribution

from trac import __version__ as TRAC_VERSION
from trac.config import Option
from trac.core import Component, implements
from trac.util.html import tag

from tracspamfilter.api import _
from tracspamfilter.captcha import ICaptchaMethod


class RecaptchaCaptcha(Component):
    """reCAPTCHA implementation"""

    implements(ICaptchaMethod)

    private_key = Option('spam-filter', 'captcha_recaptcha_private_key', '',
        """Private key for reCaptcha usage.""", doc_domain="tracspamfilter")

    public_key = Option('spam-filter', 'captcha_recaptcha_public_key', '',
        """Public key for reCaptcha usage.""", doc_domain="tracspamfilter")

    user_agent = 'Trac/%s | SpamFilter/%s' \
                 % (TRAC_VERSION, get_distribution('TracSpamFilter').version)

    def generate_captcha(self, req):
        fragment = tag.script(type='text/javascript',
                              src='https://www.google.com/recaptcha/api/'
                                  'challenge?k=%s' % self.public_key)

        # note - this is not valid XHTML!
        fragment.append(
            tag.noscript(
                tag.iframe(src='https://www.google.com/recaptcha/'
                               'api/noscript?k=%s' % self.public_key,
                           height=300, width=500, frameborder=0),
                tag.br(),
                tag.textarea(name='recaptcha_challenge_field', rows=3, cols=40),
                tag.input(type='hidden', name='recaptcha_response_field',
                          value='manual_challenge'),
                tag.br(),
                tag.input(type='submit', value=_("Submit"))
            )
        )

        return None, fragment

    def encode_if_necessary(self, s):
        if isinstance(s, unicode):
            return s.encode('utf-8')
        return s

    def verify_key(self, private_key, public_key):
        if private_key is None or public_key is None:
            return False
        # FIXME - Not yet implemented
        return True

    def verify_captcha(self, req):
        recaptcha_challenge_field = req.args.get('recaptcha_challenge_field')
        recaptcha_response_field = req.args.get('recaptcha_response_field')
        remoteip = req.remote_addr
        try:
            params = urllib.urlencode({
                'privatekey': self.encode_if_necessary(self.private_key),
                'remoteip': self.encode_if_necessary(remoteip),
                'challenge': self.encode_if_necessary(recaptcha_challenge_field),
                'response': self.encode_if_necessary(recaptcha_response_field),
            })
            request = urllib2.Request(
                url='https://www.google.com/recaptcha/api/verify',
                data=params,
                headers={
                    'Content-type': 'application/x-www-form-urlencoded',
                    'User-agent': self.user_agent
                }
            )

            response = urllib2.urlopen(request)
            return_values = response.read().splitlines()
            response.close()
        except Exception, e:
            self.log.warning("Exception in reCAPTCHA handling (%s)", e)
        else:
            if return_values[0] == 'true':
                return True
            else:
                self.log.warning("reCAPTCHA returned error: %s",
                                 return_values[1])
        return False

    def is_usable(self, req):
        return self.public_key and self.private_key



# -*- coding: utf-8 -*-
#
# Copyright (C) 2006 Edgewall Software
# All rights reserved.
#
# This software is licensed as described in the file COPYING, which
# you should have received as part of this distribution. The terms
# are also available at http://trac.edgewall.com/license.html.
#
# This software consists of voluntary contributions made by many
# individuals. For the exact contribution history, see the revision
# history and logs, available at http://projects.edgewall.com/trac/.

from trac.attachment import IAttachmentManipulator
from trac.config import IntOption
from trac.core import Component, implements
from trac.mimeview import is_binary
from trac.ticket import ITicketManipulator, TicketSystem
from trac.util.text import to_unicode
from trac.wiki.api import IWikiPageManipulator

from tracspamfilter.filtersystem import FilterSystem


class TicketFilterAdapter(Component):
    """Interface to check ticket changes for spam.
    """
    implements(ITicketManipulator)

    # ITicketManipulator methods

    def prepare_ticket(self, req, ticket, fields, actions):
        pass

    def validate_ticket(self, req, ticket):
        if 'TICKET_ADMIN' in req.perm:
            # An administrator is allowed to spam
            return []

        if 'preview' in req.args:
            # Only a preview, no need to filter the submission yet
            return []

        changes = []

        # Add the author/reporter name
        if req.authname and req.authname != 'anonymous':
            author = req.authname
        elif not ticket.exists:
            author = ticket['reporter']
        else:
            author = req.args.get('author', req.authname)

        # Add any modified text fields of the ticket
        fields = [f['name'] for f in
                  TicketSystem(self.env).get_ticket_fields()
                  if f['type'] in ('textarea', 'text')]
        for field in fields:
            if field in ticket._old:
                changes.append((ticket._old[field], ticket[field]))

        if 'comment' in req.args:
            changes.append((None, req.args.get('comment')))

        FilterSystem(self.env).test(req, author, changes)
        return []

class WikiFilterAdapter(Component):
    """Interface to check wiki changes for spam.
    """
    implements(IWikiPageManipulator)

    # IWikiPageManipulator methods

    def prepare_wiki_page(self, req, page, fields):
        pass

    def validate_wiki_page(self, req, page):
        if 'WIKI_ADMIN' in req.perm:
            # An administrator is allowed to spam
            return []

        if 'preview' in req.args:
            # Only a preview, no need to filter the submission yet
            return []

        old_text = page.old_text
        text = page.text
        author = req.args.get('author', req.authname)
        comment = req.args.get('comment')

        # Test the actual page changes as well as the comment
        changes = [(old_text, text)]
        if comment:
            changes += [(None, comment)]

        FilterSystem(self.env).test(req, author, changes)
        return []


class AttachmentFilterAdapter(Component):
    """Interface to check attachment uploads for spam.
    """
    implements(IAttachmentManipulator)

    sample_size = IntOption('spam-filter', 'attachment_sample_size', 16384,
        """The maximum number of bytes from an attachment to pass through
        the spam filters.""", doc_domain='tracspamfilter')

    # ITicketManipulator methods

    def prepare_attachment(self, req, attachment, fields):
        pass

    def validate_attachment(self, req, attachment):
        if 'WIKI_ADMIN' in req.perm:
            # An administrator is allowed to spam
            return []

        author = req.args.get('author', req.authname)
        description = req.args.get('description')

        filename = None
        upload = req.args.get('attachment')
        content = ''
        if upload is not None:
            try:
                data = upload.file.read(self.sample_size)
                if not is_binary(data):
                    content = to_unicode(data)
            finally:
                upload.file.seek(0)
            filename = upload.filename

        changes = []
        for field in filter(None, [description, filename, content]):
            changes += [(None, field)]

        FilterSystem(self.env).test(req, author, changes)
        return []



# -*- coding: utf-8 -*-
#
# Copyright (C) 2015 Dirk Stöcker <trac@dstoecker.de>
# All rights reserved.
#
# This software is licensed as described in the file COPYING, which
# you should have received as part of this distribution. The terms
# are also available at http://trac.edgewall.com/license.html.
#
# This software consists of voluntary contributions made by many
# individuals. For the exact contribution history, see the revision
# history and logs, available at http://projects.edgewall.com/trac/.

import urllib
import urllib2
from pkg_resources import get_distribution
try:
    import json
except ImportError:
    import simplejson as json

from trac import __version__ as TRAC_VERSION
from trac.config import Option
from trac.core import Component, implements
from trac.util.html import tag
from trac.web.chrome import add_script

from tracspamfilter.api import _
from tracspamfilter.captcha.api import ICaptchaMethod


class Recaptcha2Captcha(Component):
    """reCAPTCHA2 implementation"""

    implements(ICaptchaMethod)

    # use same values as reCAPTCHA1
    private_key = Option('spam-filter', 'captcha_recaptcha_private_key', '',
    """Private key for reCaptcha usage.""", doc_domain="tracspamfilter")

    public_key = Option('spam-filter', 'captcha_recaptcha_public_key', '',
        """Public key for reCaptcha usage.""", doc_domain="tracspamfilter")

    user_agent = 'Trac/%s | SpamFilter/%s' \
                 % (TRAC_VERSION, get_distribution('TracSpamFilter').version)

    def generate_captcha(self, req):
        add_script(req, 'https://www.google.com/recaptcha/api.js')

        return None, tag(
            tag.div(class_='g-recaptcha', data_sitekey=self.public_key),
            tag.input(type='submit', value=_("Submit"))
        )

    def encode_if_necessary(self, s):
        if isinstance(s, unicode):
            return s.encode('utf-8')
        return s

    def verify_key(self, private_key, public_key):
        if private_key is None or public_key is None:
            return False
        # FIXME - Not yet implemented
        return True

    def verify_captcha(self, req):
        recaptcha_response_field = req.args.get('g-recaptcha-response')
        remoteip = req.remote_addr
        try:
            params = urllib.urlencode({
                'secret': self.encode_if_necessary(self.private_key),
                'remoteip': self.encode_if_necessary(remoteip),
                'response': self.encode_if_necessary(recaptcha_response_field),
            })
            request = urllib2.Request(
                url='https://www.google.com/recaptcha/api/siteverify',
                data=params,
                headers={
                    'Content-type': 'application/x-www-form-urlencoded',
                    'User-agent': self.user_agent
                }
            )

            response = urllib2.urlopen(request)
            return_values = json.loads(response.read())
            response.close()
        except Exception, e:
            self.log.warning("Exception in reCAPTCHA handling (%s)", e)
        else:
            if return_values['success'] is True:
                return True
            else:
                self.log.warning("reCAPTCHA returned error: %s",
                                 return_values['error-codes'])
        return False

    def is_usable(self, req):
        return self.public_key and self.private_key



# -*- coding: utf-8 -*-
#
# Copyright (C) 2013 Dirk Stöcker
# All rights reserved.
#
# This software is licensed as described in the file COPYING, which
# you should have received as part of this distribution. The terms
# are also available at http://trac.edgewall.com/license.html.
#
# This software consists of voluntary contributions made by many
# individuals. For the exact contribution history, see the revision
# history and logs, available at http://projects.edgewall.com/trac/.

from __future__ import with_statement

import re

try:
    from acct_mgr.api import AccountManager
except ImportError:  # not installed
    AccountManager = None

from tracspamfilter.api import _


class UserInfo(object):

    @staticmethod
    def gettemporary(env):
        anons = []
        for sid, in env.db_query("SELECT sid FROM session WHERE authenticated = 0"):
            anons.append(sid)
        return anons

    @staticmethod
    def getuserinfo(env, username):
        data = []

        for time,page,version in env.db_query("SELECT time,name,version FROM wiki WHERE author=%s", (username,)):
            data.append((time, "/wiki/%s?version=%s" % (page,version), _("Wiki page '%(page)s' version %(version)s modified", page=page, version=version)))

        for filename,time,page,type in env.db_query("SELECT filename,time,id,type FROM attachment WHERE author=%s", (username,)):
            data.append((time, "/%s/%s#no1" % (type, page), _("Attachment '%s' added") % filename))

        for author,time,field,oldvalue,newvalue,ticket in env.db_query("SELECT author,time,field,oldvalue,newvalue,ticket FROM ticket_change"):
            if author == username:
                data.append((time, "/ticket/%s" % ticket, _("Ticket %(id)s field '%(field)s' changed", id=ticket, field=field)))
            if field == "reporter" or field == "owner":
                if oldvalue == username:
                    data.append((time, "/ticket/%s" % ticket, _("Removed from ticket %(id)s field '%(field)s' ('%(old)s' --> '%(new)s')", id=ticket, field=field, old=oldvalue, new=newvalue)))
                if newvalue == username:
                    data.append((time, "/ticket/%s" % ticket, _("Set in ticket %(id)s field '%(field)s' ('%(old)s' --> '%(new)s')", id=ticket, field=field, old=oldvalue, new=newvalue)))
            elif field == 'cc':
                authors = []
                for val in [oldvalue, newvalue]:
                    authors += UserInfo.splitcc(env, val, ticket)
                if username in authors:
                    data.append((time, "/ticket/%s" % ticket, _("Ticket %(id)s CC field change ('%(old)s' --> '%(new)s')", id=ticket, old=oldvalue, new=newvalue)))

        for time, reporter, owner, cc, id in env.db_query("SELECT time,reporter,owner,cc,id FROM ticket"):
            if reporter == username:
                data.append((time, "/ticket/%s" % id, _("Reporter of ticket %s") % id))
            if owner == username:
                data.append((time, "/ticket/%s" % id, _("Owner of ticket %s") % id))
            if username in UserInfo.splitcc(env, cc, id):
                data.append((time, "/ticket/%s" % id, _("In CC of ticket %(id)s ('%(cc)s')", id=id, cc=cc)))

        for rev,time in env.db_query("SELECT rev,time FROM revision WHERE author=%s", (username,)):
            data.append((time, None, _("Author of revision %s") % rev))

        for name, in env.db_query("SELECT name FROM component WHERE owner=%s", (username,)):
            data.append((None, "/admin/ticket/components", _("Component '%s' owner") % name))

        for sid, in env.db_query("SELECT DISTINCT(username) FROM permission WHERE username=%s", (username,)):
            data.append((None, "/admin/general/perm", _("In permissions list")))

        for id, in env.db_query("SELECT id FROM report WHERE author=%s", (username,)):
            data.append((None, "/report/%s" %id, _("Author of report %d") % id))

        # non-standard table
        try:
            for time,realm,resource in env.db_query("SELECT time,realm,resource_id FROM votes WHERE username=%s", (username,)):
                data.append((time, "/%s/%s" % (realm, resource), _("Voted for '%s'") % ("/%s/%s" % (realm, resource))))
        except: # old style
            try:
                for resource, in env.db_query("SELECT resource FROM votes WHERE username=%s", (username,)):
                    data.append((None, "/%s" % resource, _("Voted for '%s'") % resource))
            except:
                pass
        return sorted(data, key=lambda x: x[0])

    @staticmethod
    def splitcc(env, cc, ticket):
        authors = []
        if cc:
            sepchar = re.compile("[ ;]")
            for ccval in cc.split(", "):
                if ccval != '':
                    authors.append(ccval)
                    if sepchar.search(ccval):
                        env.log.warn("Strange character in CC value for "
                                     "ticket %s: '%s'", ticket, ccval)
        return authors

    @staticmethod
    def getinfo(env, mode='unused', minwiki=0):
        users = {}
        emails = {}
        emailscase = {}
        # arguments order:
        # 0: last visit time or initial link
        # 1: registered
        # 2: has settings
        # 3: e-mail
        # 4: wiki edited (1=as user, 2=as email, 3=both)
        # 5: wiki edit count
        # 6: ticket edited (1=as user, 2=as email, 3=both)
        # 7: ticket edit count
        # 8: SVN edited (only 1=as user)
        # 9: SVN edit count
        #10: component, permissions (1=as user, 2=as email, 3=both)
        #11: component, ... count
        #12: name
        #13: mail is double
        #14: is in password store

        mailre = re.compile("(?i)^(?:.*<)?([A-Z0-9._%+-]+@(?:[A-Z0-9-]+\.)+[A-Z0-9-]{2,63})(?:>.*)?$")

        def addemail(user, mail):
            finalmail = mail.lower()
            # sanitize Google mail
            m = finalmail.split("@")
            if len(m) == 2 and m[1] in ("gmail.com", "googlemail.com"):
                m = m[0].split("+") # ignore anything behind a plus
                m = m[0].replace(".", "") # ignore any dots
                finalmail = m + "@gmail.com"

            if not mail in emailscase:
                emailscase[mail] = list((user,))
            elif not user in emailscase[mail]:
                # mail already there, but longer name not, remove mail
                if mail in emailscase[mail]:
                    if users[mail][5]:
                        users[user][5] += users[sid][5]
                        users[user][4] |= 2
                    if users[mail][7]:
                        users[user][7] += users[sid][7]
                        users[user][6] |= 2
                    if users[mail][9]:
                        users[user][9] += users[sid][9]
                        users[user][8] |= 2
                    if users[mail][11]:
                        users[user][11] += users[sid][11]
                        users[user][10] |= 2
                    users.pop(mail)
                    emailscase[mail].remove(mail)
                    emails[finalmail].remove(mail)
                emailscase[mail].append(user)
            if not finalmail in emails or not len(emails[finalmail]):
                emails[finalmail] = list((user,))
            elif not user in emails[finalmail]:
                emails[finalmail].append(user)
                for u in emails[finalmail]:
                    users[u][13] = finalmail

        def getuser(mail):
            if mail in emailscase:
                return emailscase[mail][0]
            return mail

        def adduser(user, link, idx):
            if user is not None and user != "anonymous" and user != '':
                val = 1
                res = mailre.search(user)
                mail = 0
                if res:
                    mail = res.group(1)
                    if mail == user:
                        val = 2
                        user = getuser(mail)
                if not user in users:
                    users[user] = [link, 0, 0, mail, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0]
                    if mail:
                        addemail(user, mail)
                users[user][idx] |= val
                users[user][idx+1] += 1

        for sid,last_visit in env.db_query("SELECT sid,last_visit FROM session WHERE authenticated = 1"):
            users[sid] = [int(last_visit), 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0]

        for sid,name,value in env.db_query("SELECT sid,name,value FROM session_attribute WHERE authenticated = 1"):
            if not sid in users:
                users[sid] = [0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0]
            else:
                users[sid][2] = 1
            if name == "email":
                users[sid][3] = value
                addemail(sid, value)
            elif name == "name":
                users[sid][12] = value

        if AccountManager:
            for sid in AccountManager(env).get_users():
                if not sid in users:
                    users[sid] = [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1]
                else:
                    users[sid][14] += 1

        for sid,page,version in env.db_query("SELECT author,name,version FROM wiki"):
            adduser(sid, "/wiki/%s?version=%s" % (page,version), 4)

        for sid,page,type in env.db_query("SELECT author,id,type FROM attachment"):
            adduser(sid, "/%s/%s#no1" % (type, page), 6 if type == 'ticket' else 4)

        for author,field,oldvalue,newvalue,ticket in env.db_query("SELECT author,field,oldvalue,newvalue,ticket FROM ticket_change"):
            authors = [author]
            if field == "reporter" or field == "owner":
                if oldvalue is not None and oldvalue != '':
                    authors.append(oldvalue)
                if newvalue is not None and newvalue != '':
                    authors.append(newvalue)
            elif field == 'cc':
                for val in [oldvalue, newvalue]:
                    authors += UserInfo.splitcc(env, val, ticket)
            for sid in authors:
                adduser(sid, "/ticket/%s" % ticket, 6)

        for reporter, owner, cc, id in env.db_query("SELECT reporter,owner,cc,id FROM ticket"):
            authors = [reporter, owner] + UserInfo.splitcc(env, cc, id)
            for sid in authors:
                adduser(sid, "/ticket/%s" % id, 6)

        for sid, in env.db_query("SELECT author FROM revision"):
            adduser(sid, 0, 8)

        for sid, in env.db_query("SELECT owner FROM component"):
            adduser(sid, 0, 10)

        for sid, in env.db_query("SELECT username FROM permission"):
            adduser(sid, 0, 10)

        for sid, in env.db_query("SELECT author FROM report"):
            adduser(sid, 0, 10)

        # non-standard table
        try:
            for sid, in env.db_query("SELECT username FROM votes"):
                adduser(sid, 0, 10)
        except:
            pass

        killsids = []
        stats = {'numunused': 0, 'numauthorized': 0, 'numtotal': len(users)}
        for sid in users:
            if not users[sid][1]:
                if mode == 'authorized' or mode == 'unused':
                    killsids.append(sid)
            else:
                stats['numauthorized'] += 1
                wikicount = users[sid][5]
                if wikicount <= minwiki:
                    wikicount = 0
                if wikicount+users[sid][7]+users[sid][9]+users[sid][11]:
                    if mode == 'unused':
                        killsids.append(sid)
                else:
                    stats['numunused'] += 1
        for sid in killsids:
            del users[sid]
        if mode == 'overview':
            users = ()

        return users,stats

    @staticmethod
    def deletetemporary(env):
        env.db_transaction("DELETE FROM session WHERE authenticated = 0")
        env.db_transaction("DELETE FROM session_attribute WHERE authenticated = 0")

    @staticmethod
    def _fixcc(cc, old, new):
        results = []
        for entry in cc.split(", "):
            if entry == old:
                entry = new
            if not entry in results:
                results.append(entry)
        return ", ".join(results)

    @staticmethod
    def _callupdate(cursor, cmd, arg):
        cursor.execute(cmd, arg)
        try: # tested with postgres only
            return int(cursor.statusmessage[7:]) # strip "UPDATE " text
        except:
            return 0

    @staticmethod
    def changeuser(env, old, new, authorized=False):
        if authorized == "forcecc":
            forcecc = True
            authorized = None
        else:
            forcecc = False
        # prevent changing registered users
        if not authorized:
            res = env.db_query("SELECT sid,authenticated FROM session WHERE sid = %s", (old,))
            if len(res):
                return -1
        elif authorized != 'join': # already existing user
            res = env.db_query("SELECT sid,authenticated FROM session WHERE sid = %s", (new,))
            if len(res):
                return -4
        if old == new or not old or not new:
            return -2
        env.log.warn("Change username '%s' to '%s'", old, new)
        count = 0
        with env.db_transaction as db:
            cursor = db.cursor()

            entries = []
            cursor.execute("SELECT cc FROM ticket WHERE cc LIKE %s", ("%%"+old+"%%",))
            for row in cursor:
                if not row[0] in entries:
                    entries.append(row[0])

            cursor.execute("SELECT oldvalue,newvalue FROM ticket_change WHERE field='cc' AND (oldvalue LIKE %s OR newvalue LIKE %s)", ("%%"+old+"%%","%%"+old+"%%"))
            for row in cursor:
                for entry in row:
                    if entry is not None and entry != '' and \
                            entry not in entries:
                        entries.append(entry)

            sepchar = re.compile("[ ,;]")
            for entry in entries:
                newcc = UserInfo._fixcc(entry, old, new)
                if newcc != entry:
                    if sepchar.search(new) and not forcecc:
                        return -3;
                    count += UserInfo._callupdate(cursor, "UPDATE ticket SET cc = %s WHERE cc = %s", (newcc, entry))
                    count += UserInfo._callupdate(cursor, "UPDATE ticket_change SET oldvalue = %s WHERE oldvalue = %s AND field = 'cc'", (newcc, entry))
                    count += UserInfo._callupdate(cursor, "UPDATE ticket_change SET newvalue = %s WHERE newvalue = %s AND field = 'cc'", (newcc, entry))

            if authorized:
                if authorized == 'join':
                    count += UserInfo._callupdate(cursor, "DELETE FROM session WHERE sid = %s", (old,))
                    count += UserInfo._callupdate(cursor, "DELETE FROM session_attribute WHERE sid = %s", (old, ))
                else:
                    count += UserInfo._callupdate(cursor, "UPDATE session SET sid = %s WHERE sid = %s", (new, old))
                    count += UserInfo._callupdate(cursor, "UPDATE session_attribute SET sid = %s WHERE sid = %s", (new, old))
            count += UserInfo._callupdate(cursor, "UPDATE wiki SET author = %s WHERE author = %s", (new, old))
            count += UserInfo._callupdate(cursor, "UPDATE ticket_change SET author = %s WHERE author = %s", (new, old))
            count += UserInfo._callupdate(cursor, "UPDATE ticket_change SET oldvalue = %s WHERE oldvalue = %s AND (field = 'reporter' OR field = 'owner')", (new, old))
            count += UserInfo._callupdate(cursor, "UPDATE ticket_change SET newvalue = %s WHERE newvalue = %s AND (field = 'reporter' OR field = 'owner')", (new, old))
            count += UserInfo._callupdate(cursor, "UPDATE ticket SET reporter = %s WHERE reporter = %s", (new, old))
            count += UserInfo._callupdate(cursor, "UPDATE ticket SET owner = %s WHERE owner = %s", (new, old))
            count += UserInfo._callupdate(cursor, "UPDATE attachment SET author = %s WHERE author = %s", (new, old))
            count += UserInfo._callupdate(cursor, "UPDATE report SET author = %s WHERE author = %s", (new, old))
            count += UserInfo._callupdate(cursor, "UPDATE component SET owner = %s WHERE owner = %s", (new, old))
            count += UserInfo._callupdate(cursor, "UPDATE permission SET username = %s WHERE username = %s", (new, old))
            db.commit()
            # non-standard table
            try:
                try:
                    cursor.execute("DELETE from votes WHERE username = %s AND (realm, resource_id) IN (SELECT realm, resource_id FROM votes WHERE username = %s)", (old, new))
                except:
                    cursor.execute("DELETE from votes WHERE username = %s AND resource IN (SELECT resource FROM votes WHERE username = %s)", (old, new))
                count += UserInfo._callupdate(cursor, "UPDATE votes SET username = %s WHERE username = %s", (new, old))
                db.commit()
            except:
                pass

        return count



# -*- coding: utf-8 -*-
#
# Copyright (C) 2015 Edgewall Software
# Copyright (C) 2006 Alec Thomas <alec@swapoff.org>
# Copyright (C) 2015 Dirk Stöcker <trac@dstoecker.de>
# All rights reserved.
#
# This software is licensed as described in the file COPYING, which
# you should have received as part of this distribution. The terms
# are also available at http://trac.edgewall.com/license.html.
#
# This software consists of voluntary contributions made by many
# individuals. For the exact contribution history, see the revision
# history and logs, available at http://projects.edgewall.com/trac/.
#
# Author: Alec Thomas <alec@swapoff.org>

from __future__ import with_statement

import time

from trac.config import ExtensionOption, IntOption
from trac.core import Component, ExtensionPoint, Interface, TracError, \
                      implements
from trac.util.html import Markup
from trac.web.api import IRequestFilter, IRequestHandler

from tracspamfilter.api import _, IFilterStrategy, IRejectHandler, N_, \
                               RejectContent


class ICaptchaMethod(Interface):
    """ A CAPTCHA implementation. """

    def generate_captcha(req):
        """Return a tuple of `(result, html)`, where `result` is the expected
        response and `html` is a HTML fragment for displaying the CAPTCHA
        challenge."""

        """When "" is returned for result, then standard HTML code is used,
        but verification done by verify_captcha() nevertheless."""

    def verify_captcha(req):
        """Checks if captcha is valid (only in case generate_captcha returns
        None for challenge parameter, otherwise check directly."""

    def is_usable(req):
        """Check if captcha can be used for the request."""


class CaptchaSystem(Component):
    """ Main captcha handling system required to allow captcha based score updating
    in case submission was rejected.
    """
    implements(IRequestHandler, IRejectHandler, IFilterStrategy, IRequestFilter)

    handlers = ExtensionPoint(IRequestHandler)

    _name = "Captcha"

    captcha = ExtensionOption('spam-filter', 'captcha', ICaptchaMethod,
                              'ExpressionCaptcha',
        """CAPTCHA method to use for verifying humans.""",
        doc_domain="tracspamfilter")

    karma_points = IntOption('spam-filter', 'captcha_karma', 20,
        """By how many points a successful CAPTCHA response increases the
        overall score.""",
        doc_domain="tracspamfilter")

    repeat_karma_points = IntOption('spam-filter', 'captcha_failed_karma', 1,
        """By how many points a failed CAPTCHA impacts the overall score.""",
        doc_domain="tracspamfilter")

    karma_lifetime = IntOption('spam-filter', 'captcha_karma_lifetime', 86400,
        """Time in seconds that a successful CAPTCHA response increases
        karma.""", doc_domain="tracspamfilter")

    captcha_lifetime = IntOption('spam-filter', 'captcha_lifetime', 120,
        """Time in seconds before CAPTCHA is removed.""",
        doc_domain="tracspamfilter")

    captcha_cleantime = IntOption('spam-filter', 'captcha_lifetime', 3600,
        """Time in seconds before database cleanup is called.""",
        doc_domain="tracspamfilter")

    # IFilterStrategy methods

    def is_external(self):
        return False

    def _getcaptchaname(self, req):
        name = self.captcha.__class__.__name__
        if name.endswith("Captcha"):
            name = name[:-7]
        return req.session.get('captcha_verified_name', name)

    def test(self, req, author, content, ip):
        if not self._expired(req):
            self.log.debug("CAPTCHA: Test %s not expired",
                           self._getcaptchaname(req))
            return (self.karma_points,
                    N_("Human verified via CAPTCHA (%s)"),
                    self._getcaptchaname(req))
        else:
            # simply test to down-weight wrong captcha solutions
            val = int(req.session.get('captcha_reject_count', 0))
            self.log.debug("CAPTCHA: Test %s reject %d",
                           self._getcaptchaname(req), val)
            if val > 0:
                return (-self.repeat_karma_points * val,
                        N_("Failed CAPTCHA (%s) attempts"),
                        self._getcaptchaname(req))

    def train(self, req, author, content, ip, spam=True):
        pass

    # IRejectHandler methods

    def reject_content(self, req, message):
        self._cleanup()
        if self._expired(req):
            req.session['captcha_reject_time'] = int(time.time())
            val = int(req.session.get('captcha_reject_count', 0))
            req.session['captcha_reject_count'] = val + 1
            req.session['captcha_reject_reason'] = message
            req.session['captcha_redirect'] = req.path_info
            for key, value in req.args.iteritems():
                req.session['captcha_arg_%s' % key] = value
            req.redirect(req.href.captcha())
        else:
            raise RejectContent(message)

    # IRequestHandler methods

    def match_request(self, req):
        return req.path_info == '/captcha'

    def process_request(self, req):
        data = {}
        exp = req.session.get('captcha_expected')
        if req.method == 'POST':
            if req.args.get('captcha_response') == exp:
                data['error'] = _("CAPTCHA failed to handle original request")
            else:
                data['error'] = _("CAPTCHA verification failed")
        else:
            data['error'] = Markup(req.session.get("captcha_reject_reason"))
        # cleanup old values
        if exp is not None:
            del req.session['captcha_expected']
        # generate new captcha
        result, html = self.captcha.generate_captcha(req)
        data['challenge'] = html
        if self.captcha.__class__.__name__ == 'RandomCaptcha':
            data['random'] = 1
        if result == "":
            data['defaultform'] = 1
        elif result is not None:
            data['defaultform'] = 1
            req.session['captcha_expected'] = result
        req.session.save()
        return 'verify_captcha.html', data, None

    def pre_process_request(self, req, handler):
        if req.path_info == '/captcha' and req.method == 'POST':
            valid = False
            exp = req.session.get('captcha_expected')
            try:
                name = self.captcha.name(req)
            except Exception:
                name = self.captcha.__class__.__name__
            if name.endswith('Captcha'):
                name = name[:-7]
            if exp is None:
                valid = self.captcha.verify_captcha(req)
            elif req.args.get('captcha_response', '') == exp:
                valid = True

            if valid:
                req.environ['PATH_INFO'] = \
                    req.session.get('captcha_redirect', req.href())
                if 'SCRIPT_NAME' in req.environ and \
                        len(req.environ['SCRIPT_NAME']) > 1:
                    path_info = req.environ['PATH_INFO']
                    req.environ['PATH_INFO'] = \
                        path_info.replace(req.environ['SCRIPT_NAME'], '')
                req.environ['PATH_INFO'] = \
                    req.environ['PATH_INFO'].encode('utf-8')
                if 'captcha_redirect' in req.session:
                    del req.session['captcha_redirect']
                if 'captcha_reject_reason' in req.session:
                    del req.session['captcha_reject_reason']
                if 'captcha_reject_time' in req.session:
                    del req.session['captcha_reject_time']
                keys = req.session.keys()
                for key in keys:
                    if key.startswith('captcha_arg_'):
                        arg = key[12:]
                        req.args[arg] = req.session[key]
                        del req.session[key]
                try:
                    for newhandler in self.handlers:
                        try:
                            if newhandler.match_request(req):
                                keys = req.session.keys()
                                for key in keys:
                                    if key.startswith('captcha_'):
                                        self.log.info("Remove useless key: "
                                                      "%s", key)
                                        del req.session[key]
                                handler = newhandler
                                break
                        except Exception, e:
                            self.log.debug("Exception when parsing handlers: "
                                           "(%s)", e)
                except TracError, e:
                    self.log.debug("CAPTCHA: PreProcess End %s %s", name, e)
                    req.session['captcha_verified_name'] = name
                    return handler
                req.session['captcha_verified_name'] = name
                req.session['captcha_verified'] = int(time.time())
                self.log.debug("CAPTCHA: PreProcess OK %s", name)
                req.session.save()
        return handler

    def post_process_request(self, req, template, content_type):
        return template, content_type

    def post_process_request(self, req, template, data, content_type):
        return template, data, content_type

    # Internal methods

    def _expired(self, req):
        return int(req.session.get('captcha_verified', 0)) + \
               self.karma_lifetime < time.time()

    # remove old entries from database
    def _cleanup(self):
        last = 0
        for value, in self.env.db_query("""
                SELECT value FROM system WHERE name='spamfilter_lastclean'
                """):
            last = int(value)
            break
        tim = int(time.time())
        if last+self.captcha_cleantime < tim:
            self.log.debug('CAPTCHA: Cleanup captcha %s+%s < %s', last,
                           self.captcha_cleantime, tim)
            t = tim - self.captcha_lifetime
            tc = tim - self.karma_lifetime
            with self.env.db_transaction as db:
                if last == 0:
                    db("INSERT INTO system VALUES "
                       "('spamfilter_lastclean', %s)", (tim,))
                else:
                    db("UPDATE system SET value=%s WHERE "
                       "name='spamfilter_lastclean'", (tim,))
                db("""DELETE FROM session_attribute
                      WHERE
                        name LIKE 'captcha%%'
                       AND
                        (name != 'captcha_verified' OR %s < %%s)
                       AND
                        (name != 'captcha_verified_name' OR
                          (
                           sid IN
                            (SELECT * FROM
                              (SELECT sid FROM session_attribute
                                WHERE
                                  name = 'captcha_verified'
                                 AND
                                  %s < %%s
                              ) AS tmp1
                            )
                          )
                        )
                       AND
                        (
                         sid IN
                          (SELECT * FROM
                            (SELECT sid FROM session_attribute
                              WHERE
                                name = 'captcha_reject_time'
                               AND
                                %s < %%s
                            ) AS tmp1
                          )
                        OR
                         sid NOT IN
                          (SELECT * FROM
                            (SELECT sid FROM session_attribute
                              WHERE
                                name = 'captcha_reject_time'
                            ) AS tmp2
                          )
                        )
                """ % (db.cast('value', 'int'), db.cast('value', 'int'),
                       db.cast('value', 'int')), (tc, tc, t))



# -*- coding: utf-8 -*-
#
# Copyright (C) 2006 Edgewall Software
# All rights reserved.
#
# This software is licensed as described in the file COPYING, which
# you should have received as part of this distribution. The terms
# are also available at http://trac.edgewall.com/license.html.
#
# This software consists of voluntary contributions made by many
# individuals. For the exact contribution history, see the revision
# history and logs, available at http://projects.edgewall.com/trac/.

from email.Utils import parseaddr

from trac.config import IntOption
from trac.core import Component, implements

from tracspamfilter.api import IFilterStrategy, N_


class SessionFilterStrategy(Component):
    """This strategy grants positive karma points to users with an existing
    session, and extra points if they've set up their user name and password."""

    implements(IFilterStrategy)

    karma_points = IntOption('spam-filter', 'session_karma', '6',
        """By how many points an existing and configured session improves the
        overall karma of the submission. A third of the points is granted for
        having an existing session at all, the other two thirds are granted
        when the user has his name and/or email address set in the session,
        respectively.""", doc_domain="tracspamfilter")

    # IFilterStrategy implementation

    def is_external(self):
        return False

    def test(self, req, author, content, ip):
        points = 0
        if req.session.last_visit:
            points += abs(self.karma_points) / 3
            if req.session.get('name'):
                points += abs(self.karma_points) / 3
            if req.session.get('email'):
                email = parseaddr(req.session.get('email'))[1]
                if email and '@' in email:
                    points += abs(self.karma_points) / 3
            return points, N_('Existing session found')

    def train(self, req, author, content, ip, spam=True):
        return 0



# -*- coding: utf-8 -*-
#
# Copyright (C) 2006 Edgewall Software
# All rights reserved.
#
# This software is licensed as described in the file COPYING, which
# you should have received as part of this distribution. The terms
# are also available at http://trac.edgewall.com/license.html.
#
# This software consists of voluntary contributions made by many
# individuals. For the exact contribution history, see the revision
# history and logs, available at http://projects.edgewall.com/trac/.

from __future__ import with_statement

import binascii
import xml.etree.ElementTree as ElementTree
from datetime import datetime, timedelta
from time import mktime, time

from trac.db import Column, Index, Table
from trac.util.text import to_unicode

from tracspamfilter.api import gettext, _, get_strategy_name


class LogEntry(object):

    table = Table('spamfilter_log', key='id')[
        Column('id', auto_increment=True),
        Column('time', type='int'),
        Column('path'),
        Column('author'),
        Column('authenticated', type='int'),
        Column('ipnr'),
        Column('headers'),
        Column('content'),
        Column('rejected', type='int'),
        Column('karma', type='int'),
        Column('reasons'),
        Column('request')
    ]

    def __init__(self, env, time, path, author, authenticated, ipnr, headers,
                 content, rejected, karma, reasons, request):
        self.id = None
        self.env = env
        self.time = time
        self.path = path
        self.author = to_unicode(author)
        self.authenticated = bool(authenticated)
        self.ipnr = ipnr
        self.headers = to_unicode(headers) or ''
        self.content = content
        self.rejected = bool(rejected)
        self.karma = karma
        self.reasons = reasons
        self.request = request

    def __repr__(self):
        date = datetime.fromtimestamp(self.time).isoformat()
        return '<%s %s from %s by "%s">' % (self.__class__.__name__, self.id,
                                            date, self.author)

    exists = property(fget=lambda self: self.id is not None,
                      doc='Whether this log entry exists in the database')

    def _encode_content(cls, content):
        """Take a `basestring` content and return a plain text encoding."""
        return to_unicode(content).encode('utf-8').encode('base64')

    _encode_content = classmethod(_encode_content)

    def _decode_content(cls, content):
        """Revert the encoding done by `_encode_content` and return an unicode
        string"""
        try:
            return to_unicode(content.decode('base64'))
        except (UnicodeEncodeError, binascii.Error):
            # cope with legacy content (stored before base64 encoding)
            return to_unicode(content)

    _decode_content = classmethod(_decode_content)

    def get_next(self):
        """Return the next log entry in reverse chronological order (i.e. the
        next older entry.)"""
        for row in self.env.db_query("""
            SELECT id,time,path,author,authenticated,ipnr,headers,
                   content,rejected,karma,reasons,request
            FROM spamfilter_log
            WHERE id<%s ORDER BY id DESC LIMIT 1
            """, (self.id,)):
            return self.__class__._from_db(self.env, row)

    def get_previous(self):
        """Return the previous log entry in reverse chronological order
        (i.e. the next younger entry.)"""
        for row in self.env.db_query("""
            SELECT id,time,path,author,authenticated,ipnr,headers,
                   content,rejected,karma,reasons,request
            FROM spamfilter_log
            WHERE id>%s ORDER BY id LIMIT 1
            """, (self.id,)):
            return self.__class__._from_db(self.env, row)

    def insert(self):
        """Insert a new log entry into the database."""
        assert not self.exists, 'Cannot insert existing log entry'

        content = self._encode_content(self.content)
        with self.env.db_transaction as db:
            cursor = db.cursor()
            cursor.execute("""
                INSERT INTO spamfilter_log
                    (time,path,author,authenticated,ipnr,headers,content,rejected,
                     karma,reasons,request)
                VALUES (%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s)
                """, (int(self.time), self.path, self.author,
                      int(bool(self.authenticated)), self.ipnr, self.headers,
                      content, int(bool(self.rejected)), int(self.karma),
                      self._reasons_to_xml(self.reasons),
                      self._request_to_xml(self.request)))
            self.id = db.get_last_id(cursor, 'spamfilter_log')

    def update(self, **kwargs):
        """Update the log entry in the database."""
        assert self.exists, 'Cannot update a non-existing log entry'

        for name, value in kwargs.items():
            if hasattr(self, name):
                setattr(self, name, value)

        content = self._encode_content(self.content)
        self.env.db_transaction("""
            UPDATE spamfilter_log
            SET time=%s,path=%s,author=%s,
                authenticated=%s,ipnr=%s,headers=%s,content=%s,
                rejected=%s,karma=%s,reasons=%s, request=%s
            WHERE id=%s
            """, (int(self.time), self.path, self.author,
                  int(bool(self.authenticated)), self.ipnr, self.headers,
                  content, int(bool(self.rejected)), int(self.karma),
                  self._reasons_to_xml(self.reasons),
                  self._request_to_xml(self.request), self.id))

    def delete(cls, env, id):
        """Delete the log entry with the specified ID from the database."""
        if isinstance(id, list):
            env.db_transaction(
                "DELETE FROM spamfilter_log WHERE id IN (%s)"
                % ",".join("'%s'" % each for each in id))
        else:
            env.db_transaction(
                "DELETE FROM spamfilter_log WHERE id=%s", (id,))

    delete = classmethod(delete)

    def getobvious(cls, env):
        """Delete the obvious log entries from the database."""
        ids = []
        for id, reasons in env.db_query("""
            SELECT id,reasons
            FROM spamfilter_log
            WHERE authenticated=0 AND rejected=1
            """):
            for r in cls._xml_to_reasons(reasons):
                if r[0] == "Bayesian" and float(r[3]) > 90:
                    ids.append(id)
        return ids

    getobvious = classmethod(getobvious)

    def fetch(cls, env, id):
        """Retrieve an existing log entry from the database by ID."""
        for row in env.db_query("""
            SELECT id,time,path,author,authenticated,ipnr,headers,
                   content,rejected,karma,reasons,request
            FROM spamfilter_log
            WHERE id=%s
            """, (int(id),)):
            return cls._from_db(env, row)

    fetch = classmethod(fetch)

    def count(cls, env):
        """Return the number of log entries in the database."""
        for count, in env.db_query(
            "SELECT COUNT(*) FROM spamfilter_log "):
            return count

    count = classmethod(count)

    def purge(cls, env, days):
        """Delete log entries older than the specified number of days."""
        threshold = datetime.now() - timedelta(days=days)
        env.db_transaction(
            "DELETE FROM spamfilter_log WHERE time < %s",
            (mktime(threshold.timetuple()),))

    purge = classmethod(purge)

    def select(cls, env, ipnr=None, limit=None, offset=0):
        """Retrieve existing log entries from the database that match the
        specified criteria.
        """
        params = []
        where_clauses = []
        if ipnr:
            where_clauses.append("ipnr=%s")
            params.append(ipnr)

        if where_clauses:
            where = "WHERE %s" % " AND ".join(where_clauses)
        else:
            where = ""

        return cls._doselect(env, where, limit, offset, params)

    select = classmethod(select)

    def selectrelated(cls, env, page, limit=None, offset=0):
        """Retrieve page related log entries"""
        # strip additional appended values
        p = page.split("#")
        params = [p[0]]
        where_clauses = ["path = %s"]

        if page.startswith('/ticket/'):
            if len(p) == 2: # if we have time, use it
                time = p[1]
            else: # otherwise hope ticket is not yet deleted
                time = None
                for t, in env.db_query("SELECT time/1000000 from ticket WHERE id=%s", (page[8:],)):
                    time = t
            if time:
                where_clauses.append("(path = '/newticket' AND time BETWEEN %s AND %s)")
                params.append(time-30)
                params.append(time+30)

        where = "WHERE %s" % " OR ".join(where_clauses)
        return cls._doselect(env, where, limit, offset, params)

    selectrelated = classmethod(selectrelated)

    def _doselect(cls, env, where, limit, offset, params):
        extra_clauses = []

        if limit:
            extra_clauses.append("LIMIT %s")
            params.append(limit)
            if offset:
                extra_clauses.append("OFFSET %s")
                params.append(offset)
        if extra_clauses:
            extra = " ".join(extra_clauses)
        else:
            extra = ""

        for row in env.db_query("""
            SELECT id,time,path,author,authenticated,ipnr,headers,
                   content,rejected,karma,reasons,request
            FROM spamfilter_log
            %s
            ORDER BY time DESC %s
            """ % (where, extra), params):
            yield cls._from_db(env, row)

    _doselect = classmethod(_doselect)

    def _from_db(cls, env, row):
        """Create a new LogEntry from a row from the `spamfilter_log` table."""
        fields = list(row[1:])
        fields[6] = cls._decode_content(fields[6])
        fields[9] = cls._xml_to_reasons(fields[9])
        fields[10] = cls._xml_to_request(fields[10])
        obj = cls(env, *fields)
        obj.id = row[0]
        return obj

    _from_db = classmethod(_from_db)

    def _reasons_to_xml(self, reasons):
        root = ElementTree.Element("entries")
        for r in reasons:
            e = ElementTree.SubElement(root, "e", name = r[0], points = r[1], text = r[2])
            for a in r[3:]:
                ElementTree.SubElement(e, "v", value=a)
        return ElementTree.tostring(root)

    def _request_to_xml(self, request):
        if request is None:
            return None
        try:
            root = ElementTree.Element("request", target=request[0])
            for key, value in request[1].iteritems():
                ElementTree.SubElement(root, "arg", key = key).text = value
            return ElementTree.tostring(root)
        except:
            return None

    def _xml_to_reasons(self, xml):
        reasons = []
        root = ElementTree.fromstring(xml)
        for r in list(root):
            reasons.append([r.attrib.get("name"), r.attrib.get("points"), r.attrib.get("text")] +
                   [a.attrib.get("value") for a in list(r)])
        return reasons

    _xml_to_reasons = classmethod(_xml_to_reasons)

    def _xml_to_request(self, xml):
        if xml is None:
            return None
        try:
            root = ElementTree.fromstring(xml)
            request = [root.attrib.get("target"), {}]
            for r in list(root):
                request[1][r.attrib.get("key")] = r.text or ""
            return request
        except:
            return None

    _xml_to_request = classmethod(_xml_to_request)

    def getreasons(self):
        reasons = []
        for r in self.reasons:
            if len(r) == 3:
                reasons.append("%s (%s): %s" % (r[0], r[1], gettext(r[2])))
            else:
                reasons.append("%s (%s): %s" % (r[0], r[1], gettext(r[2]) % tuple(r[3:])))
        return reasons

    def findreasonvalue(self, name):
        for r in self.reasons:
            if r[0] == name:
                try:
                    return int(r[1])
                except:
                    return 0
        return 0

class Bayes(object):

    table = Table('spamfilter_bayes', key='word')[
        Column('word'),
        Column('nspam', type='int'),
        Column('nham', type='int'),
        Index(['word'])
    ]

class Statistics(object):

    table = Table('spamfilter_statistics', key=['strategy', 'action', 'data', 'status'])[
        Column('strategy'),
        Column('action'),
        Column('data'),
        Column('status'),
        Column('delay', type='double precision'),
        Column('delay_max', type='double precision'),
        Column('delay_min', type='double precision'),
        Column('count', type='int'),
        Column('external', type='int'),
        Column('time', type='int')
    ]

    def __init__(self, env):
        self.env = env

    def insert_or_update(self, strategy, action, data, status, delay, external):
        with self.env.db_transaction as db:
            row = db("SELECT delay,delay_max,delay_min,count FROM "
                "spamfilter_statistics WHERE action=%s "
                "AND data=%s AND status=%s AND strategy=%s",
                (action, data, status, strategy))
            if row:
                count = float(row[0][3])
                delay_val = float(row[0][0])*(count/(count+1.0))+delay/(count+1.0)
                delay_max = float(row[0][1])
                if delay > delay_max:
                    delay_max = delay
                delay_min = float(row[0][2])
                if delay < delay_min:
                    delay_min = delay
                count = int(row[0][3])+1
                db("UPDATE spamfilter_statistics SET delay=%s,delay_max=%s,"
                    "delay_min=%s,count=%s WHERE action=%s AND data=%s AND "
                    "status=%s AND strategy=%s", (delay_val, delay_max,
                    delay_min, count, action, data, status, strategy))
            else:
                db("INSERT INTO spamfilter_statistics VALUES (%s, %s, %s, %s, "
                    "%s, %s, %s, 1, %s, %s)", (strategy, action, data, status, delay,
                    delay, delay, external, int(time())))
        self.env.log.debug("SPAMLOG: %s %s %s %s %.3f %s", action, data,
                           status, strategy, delay,
                           "external" if external else "local")

    def clean(self, strategy):
        self.env.db_transaction(
            "DELETE FROM spamfilter_statistics WHERE strategy=%s", (strategy,))

    def cleanall(self):
        self.env.db_transaction("DELETE FROM spamfilter_statistics")

    def getstats(self):
        strategies = {}
        overall = {}
        overall['test'] = 0
        overall['time'] = 0
        overall['testspam'] = 0
        overall['testham'] = 0
        overall['testint'] = 0
        overall['testext'] = 0
        overall['testinttime'] = 0
        overall['testexttime'] = 0
        overall['spamerror'] = 0
        overall['hamerror'] = 0
        for strategy,action,data,status,delay,delay_max,delay_min,count,external,time in \
                self.env.db_query("SELECT * FROM spamfilter_statistics"):
            if strategy:
                str = strategies.get(strategy, {})
                str['type'] = "external" if external else "internal"
                str['i18type'] = _("external") if external else _("internal")
                if action == 'test':
                    if data == 'empty':
                        str['testempty'] = count
                    elif data == 'ham' or data == 'spam':
                        str['test%s%s' % (data,status)] = count
                    lasttotal = str.get('testtotal', 0)
                    str['testtotal'] = count + lasttotal
                    if lasttotal:
                        l = float(lasttotal)
                        c = float(count)
                        s = l+c
                        str['testtime'] = str['testtime']*(l/s)+float(delay)*(c/s)
                    else:
                        str['testtime'] = float(delay)
                elif action.startswith('train') or action == 'delete':
                    if status:
                        key = 'train%s%s' % (data,status)
                        str[key] = count + str.get(key, 0)
                        key = 'train%stotal' % data
                        str[key] = count + str.get(key, 0)
                        str['traintotal'] = count + str.get('traintotal', 0)
                    if action == 'trainerror' or action == 'traincond':
                        str[action] = count + str.get(action, 0)
                strategies[strategy] = str
            else:
                if not overall['time'] or time < overall['time']:
                    overall['time'] = time
                if (action == 'trainint' or action == 'trainext') and status == 'error':
                    if data == 'spam':
                        overall['spamerror'] += count
                    else:
                        overall['hamerror'] += count
                if action == 'testint' or action == 'testext':
                    overall[action] += count
                    overall["%stime" % action] = delay
                    overall['test'] += count
                    if data == 'spam':
                        overall['testspam'] += count
                    else:
                        overall['testham'] += count
        return strategies,overall

    def sortbyperformance(self, entries):
        strategies = {}
        for strategy,action,data,status,delay,count in self.env.db_query("""
            SELECT strategy,action,data,status,delay,count
            FROM spamfilter_statistics
            WHERE external=1 AND strategy != ''
            """):
            str = strategies.get(strategy, {'count': 0, 'delay': 0, 'ham': False,
                'err': 0, 'empty': 0, 'testcount': 0})
            if action == 'delete' or action.startswith('train'):
                if status == 'error':
                    str['err'] += count
                if status:
                    str['count'] += count
            elif action == 'test':
                if delay > str['delay']:
                    str['delay'] = delay
                if data == 'empty':
                    str['empty'] += count
                elif data == 'ham':
                    str['ham'] =  True
                str['testcount'] += count
            strategies[strategy] = str
        for strategy,str in strategies.iteritems():
            # 10% if ham is reported
            c = 10 if str['ham'] else 0
            # 40% if fast
            s = 40-int(str['delay']*20)
            c += s if s > 0 else 0
            # 25% if low error rate
            if str['count']:
                e = 25-int(1000.0*str['err']/str['count'])
            else:
                e = 15
            c += e if e > 0 else 0
            # 25% if high answer rate
            if str['testcount']:
                a = 25-int(25.0*str['empty']/str['testcount'])
            else:
                a = 15
            c += a if a > 0 else 0
            strategies[strategy] = c
        def mysort(val):
            return strategies.get(get_strategy_name(val), 0)
        entries = sorted(entries, key=mysort, reverse=True)
        return entries

class SpamReport(object):
    table = Table('spamfilter_report', key=['id'])[
        Column('id', auto_increment=True),
        Column('entry'),
        Column('headers'),
        Column('author'),
        Column('authenticated', type='int'),
        Column('comment'),
        Column('time', type='int')
    ]

schema = [Bayes.table, LogEntry.table, Statistics.table, SpamReport.table]
schema_version = 4



# -*- coding: utf-8 -*-
#
# Copyright (C) 2006 Edgewall Software
# Copyright (C) 2006 Alec Thomas <alec@swapoff.org>
# All rights reserved.
#
# This software is licensed as described in the file COPYING, which
# you should have received as part of this distribution. The terms
# are also available at http://trac.edgewall.com/license.html.
#
# This software consists of voluntary contributions made by many
# individuals. For the exact contribution history, see the revision
# history and logs, available at http://projects.edgewall.com/trac/.

import random

from trac.core import Component, ExtensionPoint, implements

from tracspamfilter.captcha import ICaptchaMethod


class RandomCaptcha(Component):
    """An random captcha chooser"""

    handlers = ExtensionPoint(ICaptchaMethod)

    implements(ICaptchaMethod)

    # ICaptchaMethod methods

    def generate_captcha(self, req):
        count = 0
        for handler in self.handlers:
            if not isinstance(handler, RandomCaptcha) and \
                    handler.is_usable(req):
                count += 1
        choice = random.randint(1, count)
        count = 0
        for handler in self.handlers:
            if not isinstance(handler, RandomCaptcha) and \
                    handler.is_usable(req):
                count += 1
                if count == choice:
                    req.session['captcha_handler'] = count
                    req.session.save()
                    return handler.generate_captcha(req)

    def verify_captcha(self, req):
        captcha_handler = req.session.get('captcha_handler')
        if captcha_handler is not None:
            del req.session['captcha_handler']
            captcha_handler = int(captcha_handler)
        else:
            return False
        count = 0
        for handler in self.handlers:
            if not isinstance(handler, RandomCaptcha) and \
                    handler.is_usable(req):
                count += 1
                if count == captcha_handler:
                    return handler.verify_captcha(req)
        return False

    def is_usable(self, req):
        return True

    def name(self, req):
        name = self.__class__.__name__
        captcha_handler = req.session.get('captcha_handler')
        if captcha_handler is not None:
            captcha_handler = int(captcha_handler)
            count = 0
            for newhandler in self.handlers:
                if not isinstance(newhandler, RandomCaptcha) and \
                        newhandler.is_usable(req):
                    count += 1
                    if count == captcha_handler:
                        name = newhandler.__class__.__name__
        return name



# -*- coding: utf-8 -*-
#
# Copyright (C) 2005-2012 Edgewall Software
# Copyright (C) 2005-2006 Matthew Good <trac@matt-good.net>
# Copyright (C) 2006 Christopher Lenz <cmlenz@gmx.de>
# All rights reserved.
#
# This software is licensed as described in the file COPYING, which
# you should have received as part of this distribution. The terms
# are also available at http://trac.edgewall.com/license.html.
#
# This software consists of voluntary contributions made by many
# individuals. For the exact contribution history, see the revision
# history and logs, available at http://projects.edgewall.com/trac/.
#
# Author: Matthew Good <trac@matt-good.net>
#         Christopher Lenz <cmlenz@gmx.de>

from __future__ import with_statement

import inspect
import textwrap
import time
from StringIO import StringIO
from difflib import SequenceMatcher
from pkg_resources import resource_filename

from trac.config import BoolOption, ConfigSection, ExtensionOption, IntOption
from trac.core import Component, ExtensionPoint, implements
from trac.env import IEnvironmentSetupParticipant, ISystemInfoProvider
from trac.perm import IPermissionRequestor
from trac.util import get_pkginfo
from trac.util.html import tag
from trac.util.text import shorten_line, to_unicode
from trac.web.api import Request

from tracspamfilter.api import (
    IFilterStrategy, IRejectHandler, RejectContent,
    _, N_, add_domain, get_strategy_name, gettext, tag_)
from tracspamfilter.filters.trapfield import TrapFieldFilterStrategy
from tracspamfilter.model import LogEntry, Statistics, schema_version


class FilterSystem(Component):
    """The central component for spam filtering. Must be enabled always to allow
    filtering of spam.
    """

    strategies = ExtensionPoint(IFilterStrategy)

    implements(IEnvironmentSetupParticipant, IPermissionRequestor,
               IRejectHandler, ISystemInfoProvider)

    min_karma = IntOption('spam-filter', 'min_karma', '0',
        """The minimum score required for a submission to be allowed.""",
        doc_domain='tracspamfilter')

    authenticated_karma = IntOption('spam-filter', 'authenticated_karma', '20',
        """The karma given to authenticated users, in case
        `trust_authenticated` is false.""", doc_domain='tracspamfilter')

    logging_enabled = BoolOption('spam-filter', 'logging_enabled', 'true',
        """Whether all content submissions and spam filtering activity should
        be logged to the database.""", doc_domain='tracspamfilter')

    purge_age = IntOption('spam-filter', 'purge_age', '7',
        """The number of days after which log entries should be purged.""",
        doc_domain='tracspamfilter')

    use_external = BoolOption('spam-filter', 'use_external', 'true',
        """Allow usage of external services.""", doc_domain='tracspamfilter')

    skip_external = IntOption('spam-filter', 'skip_external', '20',
        """Skip external calls when this negative karma is already reached
        by internal tests.""", doc_domain='tracspamfilter')

    skip_externalham = IntOption('spam-filter', 'skip_externalham', '30',
        """Skip external calls when this positive karma is already reached
        by internal tests.""", doc_domain='tracspamfilter')

    stop_external = IntOption('spam-filter', 'stop_external', '50',
        """Stop external calls when this negative karma is reached.""",
        doc_domain='tracspamfilter')

    stop_externalham = IntOption('spam-filter', 'stop_externalham', '50',
        """Stop external calls when this positive karma is reached.""",
        doc_domain='tracspamfilter')

    train_external = BoolOption('spam-filter', 'train_external', 'true',
        """Allow training of external services.""", doc_domain='tracspamfilter')

    trust_authenticated = BoolOption('spam-filter', 'trust_authenticated',
                                     'false',
        """Whether content submissions by authenticated users should be trusted
        without checking for potential spam or other abuse.""",
        doc_domain='tracspamfilter')

    attachment_karma = IntOption('spam-filter', 'attachment_karma', '0',
        """The karma given to attachments.""", doc_domain='tracspamfilter')

    register_karma = IntOption('spam-filter', 'register_karma', '0',
        """The karma given to registrations.""", doc_domain='tracspamfilter')

    reject_handler = ExtensionOption('spam-filter', 'reject_handler',
                                     IRejectHandler, 'FilterSystem',
        """The handler used to reject content.""", doc_domain='tracspamfilter')

    isforwarded = BoolOption('spam-filter', 'is_forwarded', 'false',
        """Interpret X-Forwarded-For header for IP checks.""",
        doc_domain='tracspamfilter')

    spam_section = ConfigSection('spam-filter',
        """This section is used to handle all configurations used by
        spam filter plugin.""", doc_domain='tracspamfilter')

    def __init__(self):
        """Set up translation domain"""
        try:
            locale_dir = resource_filename(__name__, 'locale')
        except KeyError:
            pass
        else:
            add_domain(self.env.path, locale_dir)

    # IRejectHandler methods

    def reject_content(self, req, message):
        raise RejectContent(message)

    # Public methods

    def test(self, req, author, changes):
        """Test a submission against the registered filter strategies.

        @param req: the request object
        @param author: the name of the logged in user, or 'anonymous' if the
            user is not logged in
        @param changes: a list of `(old_content, new_content)` tuples for every
            modified "field", where `old_content` may contain the previous
            version of that field (if available), and `new_content` contains
            the newly submitted content
        """

        start = time.time()
        ip = req.remote_addr
        if self.isforwarded:
            x_forwarded = req.get_header('X-Forwarded-For')
            if x_forwarded and x_forwarded != '':
                ip = x_forwarded.split(',',1)[0]

        if author.find("@") < 1:
            trap = TrapFieldFilterStrategy(self.env).get_trap(req)
            if trap:
                if trap.find("@") > 0:
                    author += " <%s>" % trap
                else:
                    self.log.debug("Append trap field to changes")
                    changes.append((None, trap))

        score = 0
        if self.trust_authenticated:
            # Authenticated users are trusted
            if req.authname and req.authname != 'anonymous':
                return

        reasons = []
        outreasons = []
        results = []
        if req.authname and req.authname != 'anonymous':
            reasons.append(("AuthenticatedUserScore", str(self.authenticated_karma),
                            N_("User is authenticated")))
            score += self.authenticated_karma

        if self.attachment_karma != 0 and \
                req.args.get('attachment') is not None:
            reasons.append(("AttachmentScore", str(self.attachment_karma),
                            N_("Attachment weighting")))
            score += self.attachment_karma

        if self.register_karma != 0 and (req.path_info == "/register" or req.args.get(TrapFieldFilterStrategy(self.env).name_register)):
            reasons.append(("RegisterScore", str(self.register_karma),
                            N_("Registration weighting")))
            score += self.register_karma

        if not author:
            author = 'anonymous'
        self.log.debug("Spam testing for %s", req.path_info)
        content = self._combine_changes(changes)
        abbrev = shorten_line(content)
        self.log.debug('Testing content %r submitted by "%s"', abbrev, author)
        externals = []

        for strategy in self.strategies:
            tim = time.time()
            try:
                if not strategy.is_external():
                    retval = strategy.test(req, author, content, ip)
                    tim = time.time()-tim
                    if tim > 3:
                        self.log.warn('Test %s took %d seconds to complete.',
                                      strategy, tim)
                    if retval:
                        points = retval[0]
                        if len(retval) > 2:
                            reason = retval[1] % retval[2:]
                        else:
                            reason = retval[1]
                        if points < 0:
                            if len(retval) > 2:
                                outreasons.append(gettext(retval[1]) % retval[2:])
                            else:
                                outreasons.append(gettext(retval[1]))

                        self.log.debug('Filter strategy %r gave submission %d '
                                       'karma points (reason: %r)', strategy,
                                       points, reason)
                        score += points
                        if reason:
                            name = get_strategy_name(strategy)
                            reasons.append((name, str(points)) + retval[1:])
                            results.append((strategy, points, tim))
                    else:
                        self._record_action('test','empty', '', strategy, tim)
                elif self.use_external:
                    externals.append(strategy)
            except Exception, e:
                self._record_action('test', 'error', '', strategy, time.time()-tim)
                self.log.exception('Filter strategy %s raised exception: %s', strategy, e)

        extint = "testint"
        if score > -self.skip_external and score < self.skip_externalham:
            externals = Statistics(self.env).sortbyperformance(externals)
            for strategy in externals:
                tim = time.time()
                try:
                    if score > -self.stop_external and score < self.stop_externalham:
                        extint = "testext"
                        retval = strategy.test(req, author, content, ip)
                        tim = time.time()-tim
                        if tim > 3:
                            self.log.warn('Test %s took %d seconds to '
                                          'complete.', strategy, tim)
                        if retval:
                            points = retval[0]
                            if len(retval) > 2:
                                reason = retval[1] % retval[2:]
                            else:
                                reason = retval[1]
                            if points < 0:
                                if len(retval) > 2:
                                    outreasons.append(gettext(retval[1]) % retval[2:])
                                else:
                                    outreasons.append(gettext(retval[1]))

                            self.log.debug('Filter strategy %r gave submission %d '
                                           'karma points (reason: %r)', strategy,
                                           points, reason)
                            score += points
                            if reason:
                                name = get_strategy_name(strategy)
                                reasons.append((name,str(points)) + retval[1:])
                                results.append((strategy, points, tim))
                        else:
                            self._record_action('test','empty', '', strategy, tim)
                except Exception, e:
                    self._record_action('test','error', '', strategy, time.time()-tim)
                    self.log.exception('Filter strategy %s raised exception: %s', strategy, e)

        reasons = sorted(reasons, key=lambda r: r[0])

        if score < self.min_karma:
            type = "spam"
            self._record_action(extint, 'spam', '', '', time.time()-start)
        else:
            type = "ham"
            self._record_action(extint, 'ham', '', '', time.time()-start)
        for strategy,points,tim in results:
            if (points < 0 and score < self.min_karma) or (points > 0 and score >= self.min_karma):
                status = "ok"
            else:
                status = "error"
            self._record_action("test", type, status, strategy, tim)

        if self.logging_enabled:
            headers = '\n'.join(['%s: %s' % (k[5:].replace('_', '-').title(), v)
                                 for k, v in req.environ.items()
                                 if k.startswith('HTTP_')])
            LogEntry(self.env, time.time(), req.path_info, author,
                     req.authname and req.authname != 'anonymous',
                     ip, headers, content, score < self.min_karma,
                     score, reasons, [req.path_info, req.args]).insert()
            LogEntry.purge(self.env, self.purge_age)

        if score < self.min_karma:
            self.log.debug('Rejecting submission %r by "%s" (%r) because it '
                           'earned only %d karma points (%d are required).',
                           abbrev, author, req.remote_addr, score, self.min_karma)
            rejects = []
            outreasons.sort()
            for r in outreasons:
                rejects.append(tag.li(r))
            msg = tag.ul(rejects)

            self.reject_handler.reject_content(
                req, tag.div(
                    tag_('Submission rejected as potential spam %(message)s',
                         message=msg),
                    class_='message'))

    def train(self, req, ids, spam=True, delete=False):
        environ = {}
        for name, value in req.environ.items():
            if not name.startswith('HTTP_'):
                environ[name] = value

        if not isinstance(ids, list):
            ids = [ids]
        for log_id in ids:
            start = time.time()
            entry = LogEntry.fetch(self.env, log_id)
            if entry:
                extint = "trainint"
                self.log.debug('Marking as %s: %r submitted by "%s"',
                               spam and 'spam' or 'ham',
                               shorten_line(entry.content),
                               entry.author)
                fakeenv = environ.copy()
                for header in entry.headers.splitlines():
                    name, value = header.split(':', 1)
                    if name == 'Cookie': # breaks SimpleCookie somehow
                        continue
                    cgi_name = 'HTTP_%s' % name.strip().replace('-', '_').upper()
                    fakeenv[cgi_name] = value.strip()
                fakeenv['REQUEST_METHOD'] = 'POST'
                fakeenv['PATH_INFO'] = entry.path
                fakeenv['wsgi.input'] = StringIO('')
                fakeenv['REMOTE_ADDR'] = entry.ipnr
                if entry.authenticated:
                    fakeenv['REMOTE_USER'] = entry.author

                type = "spam" if spam else "ham"
                for strategy in self.strategies:
                    status = "trainskip"
                    if (self.use_external and self.train_external) or not strategy.is_external():
                        tim = time.time()
                        extint = "trainext"
                        res = strategy.train(Request(fakeenv, None),
                                   entry.author or 'anonymous',
                                   entry.content, entry.ipnr, spam=spam)
                        tim = time.time()-tim
                        if tim > 3:
                            self.log.warn('Training %s took %d seconds to '
                                          'complete.', strategy, tim)
                        if res == -1:
                            status = "trainerror"
                        elif res == -2:
                            status = "traincond"
                        elif res > 0:
                            status = "train"
                    count = entry.findreasonvalue(get_strategy_name(strategy))
                    if count:
                        spamstatus = count < 0
                        self._record_action(status, type, ("ok" if spamstatus == spam else "error"), strategy, tim)
                    else:
                        self._record_action(status, type, '', strategy, tim)

                self._record_action(extint, type, ("ok" if entry.rejected == spam else "error"), '', time.time()-start)
                entry.update(rejected=spam)
                if delete:
                    self.delete(req, log_id, True)

    def delete(self, req, ids, stats):
        if not isinstance(ids, list):
            ids = [ids]
        if stats:
            for log_id in ids:
                entry = LogEntry.fetch(self.env, log_id)
                if entry:
                    type = "spam" if entry.rejected else "ham"
                    status = "delete"
                    for strategy in self.strategies:
                        count = entry.findreasonvalue(get_strategy_name(strategy))
                        if count:
                            spamstatus = count < 0
                            self._record_action(status, type, ("ok" if spamstatus == entry.rejected else "error"), strategy, 0)
                        else:
                            self._record_action(status, type, '', strategy, 0)
                    self._record_action(status, type, '', '', 0)
                    LogEntry.delete(self.env, log_id)
        else:
            LogEntry.delete(self.env,ids)

    def deleteobvious(self, req):
        ids = LogEntry.getobvious(self.env)
        if ids:
            self.delete(req, ids, True)

    # IEnvironmentSetupParticipant

    def environment_created(self):
        with self.env.db_transaction as db:
            self.upgrade_environment(db)

    def environment_needs_upgrade(self, db):
        try:
            row = db("SELECT value FROM system "
                     "WHERE name='spamfilter_version'")
            if not row or int(row[0][0]) < schema_version:
                return True
        except:
            return True

    def upgrade_environment(self, db):
        try:
            row = db("SELECT value FROM system "
               "WHERE name='spamfilter_version'")
            current_version = row and int(row[0][0]) or 0
        except:
            current_version = 0

        from tracspamfilter import upgrades
        for version in range(current_version + 1, schema_version + 1):
            for function in upgrades.version_map.get(version):
                self.log.info(textwrap.fill(inspect.getdoc(function)))
                function(self.env, db)
                self.log.info('Done.')

        if current_version == 0:
            db("INSERT INTO system VALUES ('spamfilter_version',%s)",
               (schema_version,))
            self.log.info('Created SpamFilter tables')
        else:
            db("UPDATE system SET value=%s WHERE "
               "name='spamfilter_version'", (schema_version,))
            self.log.info('Upgraded SpamFilter tables from version %d to %d',
                          current_version, schema_version)

    # IPermissionRequestor

    def get_permission_actions(self):
        return ['SPAM_CONFIG', 'SPAM_MONITOR', 'SPAM_TRAIN', 'SPAM_USER', 'SPAM_REPORT',
                'SPAM_CHECKREPORTS',
                ('SPAM_ADMIN', ['SPAM_CONFIG', 'SPAM_MONITOR', 'SPAM_TRAIN',
                'SPAM_USER', 'SPAM_REPORT', 'SPAM_CHECKREPORTS'])]

    # ISystemInfoProvider methods

    def get_system_info(self):
        # Move implementation to httpbl, ip_blacklist and url_blacklist
        # when support for Trac < 1.2 is dropped (#12294).
        try:
            import dns
        except ImportError:
            pass
        else:
            yield 'dnspython', get_pkginfo(dns)['version']

    # Internal methods

    def _combine_changes(self, changes, sep='\n\n'):
        fields = []
        for old_content, new_content in changes:
            new_content = to_unicode(new_content)
            if old_content:
                old_content = to_unicode(old_content)
                new_content = self._get_added_lines(old_content, new_content)
            fields.append(new_content)
        return sep.join(fields)

    def _get_added_lines(self, old_content, new_content):
        buf = []
        old_lines = old_content.splitlines()
        new_lines = new_content.splitlines()
        matcher = SequenceMatcher(None, old_lines, new_lines)
        for group in matcher.get_grouped_opcodes(0):
            for tag, i1, i2, j1, j2 in group:
                if tag in ('insert', 'replace'):
                    buf.append('\n'.join(new_lines[j1:j2]))

        return '\n'.join(buf)

    def _record_action(self, action, data, status, strategy, delay):
        stats = Statistics(self.env)
        if strategy:
            name = get_strategy_name(strategy)
            stats.insert_or_update(name, action, data, status, delay, 1 if strategy.is_external() else 0)
        else:
            stats.insert_or_update('', action, data, status, delay, None)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2015 Edgewall Software
# Copyright (C) 2015 Dirk Stöcker <trac@dstoecker.de>
# All rights reserved.
#
# This software is licensed as described in the file COPYING, which
# you should have received as part of this distribution. The terms
# are also available at http://trac.edgewall.com/license.html.
#
# This software consists of voluntary contributions made by many
# individuals. For the exact contribution history, see the revision
# history and logs, available at http://projects.edgewall.com/trac/.
#
# Author: Dirk Stöcker <trac@dstoecker.de>

from email.Utils import parseaddr
from pkg_resources import get_distribution
try:
    import json
except ImportError:
    import simplejson as json

from trac import __version__ as TRAC_VERSION
from trac.config import IntOption, Option, ListOption
from trac.core import Component, implements
from trac.mimeview.api import is_binary
from tracspamfilter.api import IFilterStrategy, N_
from tracspamfilter.timeoutserverproxy import TimeoutHTTPConnection
from tracspamfilter.filters.trapfield import TrapFieldFilterStrategy


class BlogSpamFilterStrategy(Component):
    """Spam filter using the BlogSpam service (http://blogspam.net/).
    """
    implements(IFilterStrategy)

    karma_points = IntOption('spam-filter', 'blogspam_karma', '5',
        """By how many points an BlogSpam reject impacts the overall
        karma of a submission.""", doc_domain='tracspamfilter')

    api_url = Option('spam-filter', 'blogspam_json_api_url',
        'test.blogspam.net:9999', "URL of the BlogSpam service.",
        doc_domain='tracspamfilter')

    skip_tests = ListOption('spam-filter', 'blogspam_json_skip_tests',
        '45-wordcount.js, 60-drone.js, 80-sfs.js',
        doc="Comma separated list of tests to skip.",
        doc_domain='tracspamfilter')

    user_agent = 'Trac/%s | SpamFilter/%s' % (
        TRAC_VERSION, get_distribution('TracSpamFilter').version
    )

    # IFilterStrategy implementation

    def is_external(self):
        return True

    def test(self, req, author, content, ip):
        if not self._check_preconditions(req, author, content):
            return

        try:
            resp = self._post(req, author, content, ip)
            if resp[0] != 200:
                raise Exception("Return code %s" % resp[0])
            resp = resp[1]
            if resp['result'] == 'SPAM':
                return (-abs(self.karma_points),
                        N_("BlogSpam says content is spam (%s [%s])"),
                        resp['reason'], resp['blocker'])
        except Exception, v:
            self.log.warning("Checking with BlogSpam failed: %s", v)

    def train(self, req, author, content, ip, spam=True):
        if not self._check_preconditions(req, author, content):
            return -2

        try:
            if spam:
                resp = self._post(req, author, content, ip, 'spam')
            else:
                resp = self._post(req, author, content, ip, 'ok')
            if resp[0] != 200:
                raise Exception("Return code %s" % resp[0])
            self.log.debug("Classifying with BlogSpam succeeded.")
            return 1
        except Exception, v:
            self.log.warning("Classifying with BlogSpam failed: %s", v)
        except IOError, v:
            self.log.warning("Classifying with BlogSpam failed: %s", v)
        return -1

    def getmethods(self):
        try:
            resp = self._call('GET', '%s/plugins' % self.api_url, None)
            if resp[0] != 200:
                raise Exception("Return code %s" % resp[0])
            return resp[1]
        except Exception, v:
            self.log.warning("Getting BlogSpam methods failed: %s", v)
            return None

    # Internal methods

    def _check_preconditions(self, req, author, content):
        if self.karma_points == 0:
            return False

        if len(content) == 0:
            return False

        if is_binary(content):
            self.log.warning("Content is binary, BlogSpam content check "
                             "skipped")
            return False

        return True

    def _post(self, req, author, content, ip, train=None):
        # Split up author into name and email, if possible
        author = author.encode('utf-8')
        author_name, author_email = parseaddr(author)
        if not author_name and not author_email:
            author_name = author
        elif not author_name and author_email.find("@") < 1:
            author_name = author
            author_email = None

        params = {
            'ip': ip,
            'name': author_name,
            'comment': content.encode('utf-8'),
            'agent': req.get_header('User-Agent'),
            'site': req.base_url,
            'version': self.user_agent
        }
        if len(self.skip_tests):
            params['options'] = 'exclude=%s' % \
                                ',exclude='.join(self.skip_tests)
        if author_email:
            params['email'] = author_email
        trap_link = TrapFieldFilterStrategy(self.env).getlink(req)
        if trap_link:
            params['link'] = trap_link

        if train is not None:
            params['train'] = train
            return self._call('POST', '%s/classify' % self.api_url, params)
        else:
            return self._call('POST', '%s/' % self.api_url, params)

    def _call(self, method, url, data=None):
        """ Do the actual HTTP request """
        offs = url.find('/')
        api_host = url[:offs]
        path = url[offs:]
        conn = TimeoutHTTPConnection(api_host)
        headers = {'User-Agent': self.user_agent}

        if data:
            headers.update({'Content-type': 'application/json'})
            conn.request(method, path, json.dumps(data), headers)
        else:
            conn.request(method, path, None, headers)

        response = conn.getresponse()
        body = response.read()
        body = json.loads(body)
        result = [response.status, body]
        conn.close()
        return result



# -*- coding: utf-8 -*-
#
# Copyright (C) 2006 Edgewall Software
# All rights reserved.
#
# This software is licensed as described in the file COPYING, which
# you should have received as part of this distribution. The terms
# are also available at http://trac.edgewall.com/license.html.
#
# This software consists of voluntary contributions made by many
# individuals. For the exact contribution history, see the revision
# history and logs, available at http://projects.edgewall.com/trac/.


import unittest

from trac.test import EnvironmentStub, Mock

from tracspamfilter.filters.session import SessionFilterStrategy


class SessionFilterStrategyTestCase(unittest.TestCase):

    def setUp(self):
        self.env = EnvironmentStub(enable=[SessionFilterStrategy])
        self.strategy = SessionFilterStrategy(self.env)

    def test_new_session(self):
        data = {}
        session = Mock(last_visit=42, get=data.get)
        rv = self.strategy.test(Mock(session=session), None, None,
                                '127.0.0.1')
        self.assertEqual((2, "Existing session found"), rv)

    def test_session_name_set(self):
        data = {'name': 'joe'}
        session = Mock(last_visit=42, get=data.get)
        rv = self.strategy.test(Mock(session=session), None, None,
                                '127.0.0.1')
        self.assertEqual((4, "Existing session found"), rv)

    def test_session_email_set(self):
        data = {'email': 'joe@example.org'}
        session = Mock(last_visit=42, get=data.get)
        rv = self.strategy.test(Mock(session=session), None, None,
                                '127.0.0.1')
        self.assertEqual((4, "Existing session found"), rv)

    def test_session_email_set_but_invalid(self):
        data = {'email': 'joey'}
        session = Mock(last_visit=42, get=data.get)
        rv = self.strategy.test(Mock(session=session), None, None,
                                '127.0.0.1')
        self.assertEqual((2, "Existing session found"), rv)

    def test_session_name_and_email_set(self):
        data = {'name': 'joe', 'email': 'joe@example.org'}
        session = Mock(last_visit=42, get=data.get)
        retval = self.strategy.test(Mock(session=session), None, None,
                                    '127.0.0.1')
        self.assertEqual((6, "Existing session found"), retval)


def test_suite():
    suite = unittest.TestSuite()
    suite.addTest(unittest.makeSuite(SessionFilterStrategyTestCase))
    return suite

if __name__ == '__main__':
    unittest.main(defaultTest='test_suite')



# -*- coding: utf-8 -*-
#
# Copyright (C) 2012 Dirk Stöcker <trac@dstoecker.de>
# All rights reserved.
#
# This software is licensed as described in the file COPYING, which
# you should have received as part of this distribution. The terms
# are also available at http://trac.edgewall.com/license.html.
#
# This software consists of voluntary contributions made by many
# individuals. For the exact contribution history, see the revision
# history and logs, available at http://projects.edgewall.com/trac/.

import string
import urllib2
from email.Utils import parseaddr
from pkg_resources import get_distribution
from urllib import urlencode

from trac import __version__ as TRAC_VERSION
from trac.config import IntOption, Option
from trac.core import Component, implements

from tracspamfilter.api import IFilterStrategy, N_


class BotScoutFilterStrategy(Component):
    """Spam filter using the BotScout (http://botscout.com/).
    """
    implements(IFilterStrategy)

    karma_points = IntOption('spam-filter', 'botscout_karma', '3',
        """By how many points a BotScout reject impacts the overall karma of
        a submission.""", doc_domain='tracspamfilter')

    api_key = Option('spam-filter', 'botscout_api_key', '',
        "API key required to use BotScout.", doc_domain='tracspamfilter')

    user_agent = 'Trac/%s | SpamFilter/%s' % (
        TRAC_VERSION, get_distribution('TracSpamFilter').version
    )

    # IFilterStrategy implementation

    def is_external(self):
        return True

    def test(self, req, author, content, ip):
        if not self._check_preconditions(False):
            return
        try:
            resp = self._send(req, author, ip)
            if resp.startswith('Y'):
                count = 0
                res = string.split(resp, '|')
                if res[3] != '0':
                    count += 1
                if res[5] != '0':
                    count += 1
                if res[7] != '0':
                    count += 1
                return (-abs(self.karma_points) * count,
                        N_("BotScout says this is spam (%s)"), resp)
        except urllib2.URLError, e:
            self.log.warn("BotScout request failed (%s)", e)

    def train(self, req, author, content, ip, spam=True):
        return 0

    # Internal methods

    def _check_preconditions(self, train):
        if self.karma_points == 0:
            return False

        if not self.api_key:
            return False

        return True

    def _send(self, req, author, ip):
        # Split up author into name and email, if possible
        author = author.encode('utf-8')
        author_name, author_email = parseaddr(author)
        if not author_name and not author_email:
            author_name = author
        elif not author_name and author_email.find('@') < 1:
            author_name = author
            author_email = None
        if author_name == 'anonymous':
            author_name = None

        params = {'ip': ip, 'key': self.api_key}
        if author_name:
            params['name'] = author_name
        if author_email:
            params['mail'] = author_email

        url = 'http://botscout.com/test/?multi&' + urlencode(params)
        url_req = urllib2.Request(url, None, {'User-Agent': self.user_agent})

        resp = urllib2.urlopen(url_req)
        return resp.read()



# -*- coding: utf-8 -*-
#
# Copyright (C) 2006 Edgewall Software
# Copyright (C) 2006 Alec Thomas <alec@swapoff.org>
# All rights reserved.
#
# This software is licensed as described in the file COPYING, which
# you should have received as part of this distribution. The terms
# are also available at http://trac.edgewall.com/license.html.
#
# This software consists of voluntary contributions made by many
# individuals. For the exact contribution history, see the revision
# history and logs, available at http://projects.edgewall.com/trac/.
#
# Author: Alec Thomas <alec@swapoff.org>

import random

from trac.core import Component, TracError, implements
from trac.config import IntOption

from tracspamfilter.captcha import ICaptchaMethod
from tracspamfilter.api import N_, _, gettext


class ExpressionCaptcha(Component):
    """captcha in the form of a human readable numeric expression

    Initial implementation by sergeych@tancher.com.
    """

    implements(ICaptchaMethod)

    terms = IntOption('spam-filter', 'captcha_expression_terms', 3,
            """Number of terms in numeric CAPTCHA expression.""",
            doc_domain='tracspamfilter')

    ceiling = IntOption('spam-filter', 'captcha_expression_ceiling', 10,
            """Maximum value of individual terms in numeric CAPTCHA
            expression.""", doc_domain='tracspamfilter')

    operations = {'*': N_("multiplied by"), '-': N_("minus"),
                  '+': N_("plus")}

    numerals = (N_("zero"), N_("one"), N_("two"), N_("three"),
                N_("four"), N_("five"), N_("six"), N_("seven"),
                N_("eight"), N_("nine"), N_("ten"), N_("eleven"),
                N_("twelve"), N_("thirteen"), N_("fourteen"),
                N_("fifteen"), N_("sixteen"), N_("seventeen"),
                N_("eighteen"), N_("nineteen"))

    # TRANSLATOR: if compound numbers like in english are not
    # supported, simply add a "plus" command to the following
    # translations!
    tens = (N_("twenty"), N_("thirty"), N_("forty"), N_("fifty"),
            N_("sixty"), N_("seventy"), N_("eighty"), N_("ninety"))

    # ICaptchaMethod methods

    def generate_captcha(self, req):
        if self.ceiling > 100:
            raise TracError(
                _("Numeric captcha can not represent numbers > 100"))
        terms = [unicode(random.randrange(0, self.ceiling))
                 for i in xrange(self.terms)]
        operations = [random.choice(self.operations.keys())
                      for i in xrange(self.terms)]
        expression = sum(zip(terms, operations), ())[:-1]
        expression = eval(compile(' '.join(expression), 'captcha_eval', 'eval'))
        human = sum(zip([self._humanise(int(t)) for t in terms],
                        [gettext(self.operations[o]) for o in operations]),
                    ())[:-1]
        return expression, u' '.join(map(unicode, human))

    def verify_captcha(self, req):
        return False

    def is_usable(self, req):
        return True

    # Internal methods

    def _humanise(self, value):
        if value < 20:
            return gettext(self.numerals[value])
        english = gettext(self.tens[value / 10 - 2])
        if value % 10:
            english += u' ' + gettext(self.numerals[value % 10])
        return english
